/*!

	\file		_aos.h
	\brief	Private library functions of the AOS (Attribute Object Store)
	\author	Aram Verstegen

	This header contains some private functions of the Attribute Object Store.
	It contains the following functions:
  -# AOS_buffer_size(arg_type): returns size of buffer needed to store an argument of arg_type;
  -# AOS_free_argslist(void): Free's linkedlist nodes that have malloc-ed buffers.
  -# AOS_get_last_node(void): Returns last node in the linkedlist.
  -# AOS_is_initialized(void): Returns 1 if the AOS list is initialized, 0 if it isn't. This might be deprecated now that the AOS uses a linked list rather than an array.
*/

#ifndef _AOS_H
#define _AOS_H
#if HAVE_STRING_H
#include <string.h>
#endif
#if HAVE_STDIO_H
#include <stdio.h>
#endif
#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if HAVE_ERRNO_H
#include <errno.h>
#endif
#if HAVE_PTHREAD_H
#include <pthread.h>
#endif
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#include "eef_log.h"
#include "eef_return_codes.h"
#include "eef_plugindl_s.h"
#include "eef_polytypes.h"

/**
 * \struct aos_attribute_s
 * A struct representing an attribute containing arbitrary data.
 */
struct aos_attribute_s {
  char *                      id;                              /*! a label to identify the data */
  char *                      issuer;                          /*! a string to identify the issuer */
  char *                      type;                            /*! a type identifier */
  int                         needs_free;                      /*! a flag that determines wheter the data should be free'd */
  void*                       data;                            /*! a pointer to store the data */
  struct aos_attribute_s *    parent;                          /*! a pointer to the 'parent' object */
  struct aos_attribute_s *    child;                           /*! a pointer to the 'child' object */
  struct aos_attribute_s *    next;                            /*! a pointer to the next node */
  eef_plugindl_t *            setting_plugin;                  /*! a pointer to the plugin that initialized this data */
};

/**
 * \struct aos_context_s
 * A container for attributes of a specific class.
 */
struct aos_context_s {
  struct aos_attribute_s*           list_attributes;                 /** a list of attributes */
  struct aos_attribute_s*           last_attribute;                  /** the last attribute in the list */
  enum aos_context_class_e          context_class;                   /** an aos_context_class_e denoting the type of attributes contained here */
  char*                             obligation_name;                 /** used to store the name of an obligation if this context is of the class OBLIGATION */
  struct aos_context_s *            next;                            /** pointer to the next context in the list */
};

/**
 * \struct aos_storage_s
 * A container for a list of contexts.
 */
struct aos_storage_s {
  struct aos_context_s*             list_contexts;             /** a list of contexts */
};

struct aos_state_s {
  struct aos_storage_s*             current_storage;
  struct aos_storage_s*             saved_storage;
  struct aos_context_s*             last_context;              /** the last context in the list */
};

EES_RC                              aos_free_storage(struct aos_storage_s*);
EES_RC                              aos_free_state(struct aos_state_s* state);
int                                 aos_is_initialized(void);
struct aos_storage_s *              aos_get_storage(void);
void                                aos_make_key(void);
void                                aos_clean_key(void *v);
void                                aos_free_key(void* storage);

EES_RC                              aos_set_iterator(struct aos_context_s* context);
struct aos_context_s*               aos_get_iterator(void);
struct aos_state_s*                 aos_get_state(void);
/*struct aos_storage_s*               localize_storage(struct aos_storage_s* storage);*/

#endif

