/*!

        \file           pdl.h
        \brief  Callback functions for parser of the evaluation manager
        \author Aram Verstegen

        This header contains some callback functions used for the PDL parser:
        It contains the following functions:

  -# pdl_init(const char*): Inilizes the parser
  -# pdl_term(void): terminates the parser
  -# set_modules_path(record_t*): sets the path to use for plug-ins (used by the parser)
  -# get_pdl_path(void): returns the path read from the config file or empty string if none was parsed
  -# add_policy(record_t*, rule_t*): adds a new policy
  -# add_rule(record_t*, record_t*, record_t*): adds a new rule
  -# remove_policy(record_t* policy): removes a policy
  -# add_variable(record_t* name, record_t* value): adds a variable (plug-in) to the list
  -# allow_rules(int):
  -# allowed_policy_rule(const char* label):

  -# concat_strings(record_t*, record_t*): concatenates two strings
  -# concat_strings_with_space(record_t*, record_t*): concatenates two strings separated by a space
 */

#ifndef PDL_H
#define PDL_H
#if HAVE_STDIO_H
#include <stdio.h>
#endif
#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#include "eef_polytypes.h"
#include "eef_plugindl_s.h"
#include "eef_private_types.h"
#include "eef_log.h"

#define TRUE 1
#define FALSE 0

/* Policy parser types */

/**
 * \struct em_record_s
 * \brief Structure for symbols used by the parser internally
 */
typedef struct em_record_s {
  char*               string;       /*!<  Hold the symbol that lex has found. */
  int                 lineno;       /*!<  Hold the line number the symbol has been found. */
} record_t;


/**
 * \struct em_var_s
 * \brief Structure keeps track of the variables, their value and the line number they are defined on.
 */
typedef struct em_var_s {
  char*               name;         /*!<  Name of the variable. */
  char*               value;        /*!<  Value of the variable. */
  unsigned int        lineno;       /*!<  Line number the variable appears on. */
  struct em_var_s*    next;         /*!<  Next variable, or 0 if none. */
} var_t;

/**
 * \struct em_rule_s
 * \brief Structure to make a tree of 
 */
typedef struct em_rule_s {
  unsigned int        lineno;       /*!<  Line number where rule appeared. */
  char*               name;         /*!<  Name of the state.*/
  struct em_rule_s*   true_branch;  /*!<  True branch or NULL if none */
  struct em_rule_s*   false_branch; /*!<  False branch or NULL if none */
  struct em_rule_s*   next;         /*!<  Next rule, or NULL if none */
  eef_plugindl_t*     plugin;
} rule_t;

/**
 * \struct em_policy_s
 * \brief Structure to build a tree of plugins as defined in a policy in the config file
 */
typedef struct em_policy_s {
  char*               name;
  unsigned int        lineno;
  rule_t*             rules;
  int                 rules_list_transformed_to_tree;
  struct em_policy_s* next;
} policy_t;

/**
 * \struct pdl_data_s
 * A struct that holds internal data, and some error flags.
 */
typedef struct pdl_data_s {
  const char* config_file_s; 
  FILE*       config_file_fp;
  char*       pdl_path;

  int         pdl_port; 

  char*       pdl_log_file;

  int         recursion_was_created;
  int         unknown_variable_was_referenced;
  int         starting_state_was_reused;
  int         parse_errors_detected;

  var_t*      variables_list;
  var_t*      variables_list_last;
  var_t*      current_variable;

  rule_t*     rules_list;
  rule_t*     rules_list_last;

  policy_t*   policies_list;
  policy_t*   policies_list_last;
} pdl_data_t;


/* lifecycle functions */
EES_RC              start_pdl_parser(char *config_file, int number_of_policies, char* names_of_policies[]);
EES_RC              wrap_yacc(const char* config_file);
EES_RC              pdl_term(void);
EES_RC              add_plugin_structs(void);

policy_t*           get_policies(void);
policy_t*           reduce_policies(policy_t* policies, int number_of_policies, char* names_of_policies[]);
EES_RC              link_rules_to_plugins(policy_t* policies);
void                link_rule_to_plugin(policy_t *policy, rule_t *rule);

EES_RC              clean_variables_list(void);
EES_RC              clean_policies_list(policy_t* policies);
rule_t*             clean_rules_list(rule_t*);
rule_t*             clean_rules_tree(rule_t*);

/* callback functions */
void                add_policy(record_t*, rule_t*);
rule_t *            add_rule(record_t*, record_t*, record_t*);

void                set_pdl_path(record_t*);
const char *        get_pdl_path(void);

void                set_pdl_log_file(record_t*);
const char *        get_pdl_log_file(void);

int                 check_path(char* path);

void                remove_policy(record_t* policy);
policy_t*           remove_policy_by_name(policy_t* policies, char* policy);
void                add_variable(record_t* name, record_t* value);
EES_RC              allow_rules(int);
EES_RC              allowed_policy_rule(const char* label);
void                set_pdl_port(record_t*);
int                 get_pdl_port(void);

/* convenience methods used by the lexer */
record_t *          concat_strings(record_t*, record_t*);
record_t *          concat_strings_with_space(record_t*, record_t*);
char **             _var_to_argv(char*, int*);
var_t*              get_variable_by_name(char*);
rule_t*             get_left_hand_rule(rule_t*, const char*);
rule_t*             get_right_hand_rule(rule_t*, const char*);
rule_t*             check_for_recursion(rule_t*, rule_t*);
void                free_args(int argc, char** argv);
void                link_dead_end_rules_in_policies(policy_t* policy);
void                link_dead_end_rules(policy_t* policy, rule_t* rule);
EES_RC              remove_unreachable_rules_in_policies(policy_t* policies);
rule_t*             remove_unreachable_rules(rule_t* rules, EES_RC*);

/* Lex/Yacc stuff */
int                 yyerror(char*);
void                set_yylval(record_t* r);
#if HAVE_FLEX
void                delete_lex_buffer(void);
#endif
#if 0
/* Note: flex will set a prototype when needed */
#if HAVE_YYLEX_DESTROY
extern int          yylex_destroy(void);
#endif
#endif


/* debug functions */
void                print_policies(policy_t* policies);
void                print_rules(policy_t* policy, rule_t* rules);

#endif
