#define _GNU_SOURCE
#include "eef_plugin.h"
#include "gridmapfile.h"
#include "gridmapdir.h"
#include "unixpriv.h"
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <pwd.h>
#include <sys/stat.h>

EES_PL_RC set_var_as_string(const char* name, char* value);
EES_PL_RC set_var_as_int(const char* name, int value);
EES_PL_RC set_var_as_void_p(const char* name, void* value, size_t size);
char* get_var_as_string(char* name);
int get_var_as_int(char* name);
void* get_var_as_void_p(char* name);
char* dn_to_username(char* _dn);
char* find_in_mappings(const char* string, gridmap_list_t *_mappings);

EES_PL_RC plugin_initialize(int argc, char* argv[]){
  int               option_index, c;    /* getopt */

  char*             _obligation_name   = NULL;   /* obligation name to create  */
  char*             _attribute_user    = NULL;   /* attribute name for user */
  char*             _attribute_group   = NULL;   /* attribute name for group */

  char*             _gridmapfile       = NULL;   /* path to gridmapfile */
  char*             _gridmapdir        = NULL;   /* path to gridmapdir */

  struct stat       _gridmapfile_stat;           /* stat() structs to check existence of gridmapfile */
  struct stat       _gridmapdir_stat;            /* stat() structs to check existence of gridmapdir */

  char*             _mapping_attribute = NULL;   /* which field to retrieve DN from */

  int               _use_usernames     = 0;      /* wheter to use username strings. uses uid strings otherwise */
  int               _use_pool          = 0;      /* use pool accounts from gridmapdir */

  int               _retval            = 0;
  gridmap_list_t   *_mappings          = NULL;

  static struct option long_options[] = /* options */
  {
    /*{"gridmap",       required_argument, 0, 'g'},*/
    {"gridmapfile",   required_argument, 0, 'f'},
    {"gridmapdir" ,   required_argument, 0, 'd'},
    {"mapattrib",     required_argument, 0, 'm'},
    {"obligation",    required_argument, 0, 'o'},
    {"attribute_u",   required_argument, 0, 'u'},
    {"attribute_g",   required_argument, 0, 'g'},
    {"strings",       no_argument,       0, 'n'},
    {"pool",          no_argument,       0, 'p'}
  };

  EEF_log(LOG_DEBUG, "%s: Initializing localaccount plugin!\n", EEF_getRunningPluginName());

  opterr = 0;
  optind = 0;

  /* parse options */
  while(1){
    c = getopt_long_only(argc, argv, "", long_options, &option_index);
    if(c == -1){
      break;
    }
    switch(c){
      case 'f':
        if((_gridmapfile = strdup(optarg)) == NULL){
          EEF_log(LOG_ERR, "%s: %s", EEF_getRunningPluginName(), strerror(errno)); 
        }
        break;
      case 'd':
        if((_gridmapdir = strdup(optarg)) == NULL){
          EEF_log(LOG_ERR, "%s: %s", EEF_getRunningPluginName(), strerror(errno)); 
        }
        break;
      case 'm':
        if((_mapping_attribute = strdup(optarg)) == NULL){
          EEF_log(LOG_ERR, "%s: %s", EEF_getRunningPluginName(), strerror(errno)); 
        }
        break;
      case 'o':
        if((_obligation_name = strdup(optarg)) == NULL){
          EEF_log(LOG_ERR, "%s: %s", EEF_getRunningPluginName(), strerror(errno)); 
        }
        break;
      case 'u':
        if((_attribute_user = strdup(optarg)) == NULL){
          EEF_log(LOG_ERR, "%s: %s", EEF_getRunningPluginName(), strerror(errno)); 
        }
        break;
      case 'g':
        if((_attribute_group = strdup(optarg)) == NULL){
          EEF_log(LOG_ERR, "%s: %s", EEF_getRunningPluginName(), strerror(errno)); 
        }
        break;
      case 'n':
        _use_usernames = 1;
        break;
      case 'p':
        _use_pool = 1;
        break;
      case '?':
        EEF_log(LOG_ERR, "%s: unknown option %s", EEF_getRunningPluginName(), optarg);
        break;
      case ':':
        EEF_log(LOG_ERR, "%s: missing argument for %s", EEF_getRunningPluginName(), optarg);
        break;
    }
  }

  if(_gridmapfile){
    /* check if path to gridmapfile exists */
    if(stat(_gridmapfile, &_gridmapfile_stat) != 0){
      EEF_log(LOG_ERR, "Unable to stat gridmapfile (%s): %s\n", _gridmapfile, strerror(errno));
      return EES_PL_FAILURE;
    }
  }

  if(_gridmapdir){
    /* check if path to gridmapdir exists */
    if(stat(_gridmapdir, &_gridmapdir_stat) != 0){
      EEF_log(LOG_ERR, "Unable to stat gridmapdir (%s): %s\n", _gridmapdir, strerror(errno));
      return EES_PL_FAILURE;
    }
  }

  _retval = parse_gridmapfile(_gridmapfile, &_mappings);

  /* look up DN in gridmapfile */
  if(_retval == 0){
    set_var_as_void_p("gridmappings", &_mappings, sizeof(void*));
  } else {
    switch(_retval){
      case -1:
        EEF_log(LOG_ERR, "%s: Unable to parse gridmapfile: %s - I/O error (file may have changed during access)\n", EEF_getRunningPluginName(), _gridmapfile);
        break;
      case -2:
        EEF_log(LOG_ERR, "%s: Unable to parse gridmapfile: %s - privilege-drop error\n", EEF_getRunningPluginName(), _gridmapfile);
        break;
      case -3:
        EEF_log(LOG_ERR, "%s: Unable to parse gridmapfile: %s - permission error (untrusted path)\n", EEF_getRunningPluginName(), _gridmapfile);
        break;
      case -4:
        EEF_log(LOG_ERR, "%s: Unable to parse gridmapfile: %s - memory error \n", EEF_getRunningPluginName(), _gridmapfile);
        break;
      case -5:
        EEF_log(LOG_ERR, "%s: Unable to parse gridmapfile: %s - unknown or safefile error \n", EEF_getRunningPluginName(), _gridmapfile);
        break;
      default:
        EEF_log(LOG_ERR, "%s: Unable to parse gridmapfile: %s - mysterious unknown error \n", EEF_getRunningPluginName(), _gridmapfile);
        break;
    }
    return EES_PL_FAILURE;
  }

  /* save state to AOS */
  set_var_as_string("user_attribute",  _attribute_user);
  set_var_as_string("group_attribute", _attribute_group);
  set_var_as_string("obligation_name", _obligation_name);
  set_var_as_string("gridmapfile",     _gridmapfile);
  set_var_as_string("gridmapdir",      _gridmapdir);
  set_var_as_string("mappingattr",     _mapping_attribute);
  set_var_as_int("use_pool",           _use_pool);
  set_var_as_int("use_usernames",      _use_usernames);
  aos_dump_argslist();

  return EES_PL_SUCCESS;
}

EES_PL_RC plugin_run(){
  aos_context_t    *_context            = NULL;
  aos_attribute_t  *_tmp_attr           = NULL;

  aos_attribute_t  *_uid_attr           = NULL;
  aos_attribute_t  *_gid_attr           = NULL;

  char             *_dn                 = NULL;

  char             *_uid_str            = NULL;
  char             *_gid_str            = NULL;
  int		    uid_str_len;

  char             *_user_name          = NULL;
  char             *_attribute_user     = NULL;
  char             *_attribute_group    = NULL;
  char             *_obligation_name    = NULL;

  /*char             *_gridmapdir         = NULL;*/

  char             *_mapping_attribute  = NULL;
  int               _use_usernames      = 0;
  int               _use_pool           = 0;
  char             *_cmp_attr_id        = NULL;
  /*struct passwd     _tmp_pwd_s;*/

  EEF_log(LOG_INFO, "Running %s\n", EEF_getRunningPluginName());

  /* populate fields */
  _attribute_user    = get_var_as_string("user_attribute");
  _attribute_group   = get_var_as_string("group_attribute");
  _obligation_name   = get_var_as_string("obligation_name");
  _mapping_attribute = get_var_as_string("mappingattr");

  _use_usernames     = get_var_as_int("use_usernames");
  _use_pool          = get_var_as_int("use_pool");

  /* fetch DN (subject-x509-id) */
  rewindContexts();
  while((_context = getNextContext(SUBJECT, NULL))){
    rewindAttributes(_context);
    while((_tmp_attr = getNextAttribute(_context))){
      _cmp_attr_id = getAttributeId(_tmp_attr);
      if(_cmp_attr_id && _mapping_attribute){
        if(strcmp(_cmp_attr_id, _mapping_attribute) == 0){
          _dn = getAttributeValueAsString(_tmp_attr);
        }
      }
    }
  }

  aos_dump_argslist();

  /* check if a DN was found */
  if(_dn == NULL){
    EEF_log(LOG_ERR, "%s: No DN found!\n", EEF_getRunningPluginName());
    return EES_PL_FAILURE;
  } else {
    _user_name = dn_to_username(_dn);
  }

  if(_user_name == NULL){
    EEF_log(LOG_ERR, "%s: Unable to find local mapping for DN \"%s\"\n", EEF_getRunningPluginName(), _dn);
    return EES_PL_FAILURE;
  } else {
    EEF_log(LOG_ERR, "%s: Found mapping for DN \"%s\": account \"%s\"\n", EEF_getRunningPluginName(), _dn, _user_name);
  }

  /* TODO add secondary GIDs */

  if(_use_usernames){
    /* username as string */
    _uid_str = _user_name;
  } else {
    /* username as uid */
/*    asprintf(&_uid_str, "%llu", (long long unsigned int) uid_from_user_name(_user_name));*/

    uid_str_len=sizeof(uid_t)*3+2; /* 3 for 8*log(2)/log(10), 2 for -1 and \0 */
    if ((_uid_str=(char*)malloc(uid_str_len))==NULL)  {
	EEF_log(LOG_ERR, "out of memory\n");
        return EES_PL_FAILURE;
    }
    snprintf(_uid_str, uid_str_len, "%u", (int)threadsafe_getuid_from_name(_user_name));
   
  }

  EEF_log(LOG_INFO, "Adding obligations for uid %s and gid %s\n", _uid_str, _gid_str);

  /* add obligation for new uid-gid */

  if((_context = createContext(OBLIGATION)) != NULL){
    setContextObligationId(_context, _obligation_name);
    if((_uid_attr = createAttribute())){
      setAttributeId(   _uid_attr, _attribute_user);
      setAttributeValue(_uid_attr, _uid_str, strlen(_uid_str)+1);
      addAttribute(_context, _uid_attr);
      free(_uid_str);
    }

    if(_gid_str){
      if((_gid_attr = createAttribute())){
        setAttributeId(   _gid_attr, _attribute_group);
        setAttributeValue(_gid_attr, _gid_str, strlen(_gid_str)+1);
        addAttribute(_context, _gid_attr);
      }
      free(_gid_str);
    }

    addContext(_context);
    return EES_PL_SUCCESS;
  }

  return EES_PL_FAILURE;
}

char* dn_to_username(char* _dn){
  char *_gridmapfile        = get_var_as_string("gridmapfile");
  char *_gridmapdir         = get_var_as_string("gridmapdir");
  int   _use_pool           = get_var_as_int("use_pool");
  char *_single_user_name   = NULL;
  char* _pool_user_name     = NULL;
  gridmap_list_t  *_mappings       = NULL;

  _mappings = get_var_as_void_p("gridmappings");
  _gridmapfile = get_var_as_string("gridmapfile");
  EEF_log(LOG_DEBUG, "Get gridmappings into AOS at %p\n", _mappings);
  _single_user_name = find_in_mappings(_dn, _mappings);

  if(_use_pool && _single_user_name){
    /* get pool account from gridmapdir */
    EEF_log(LOG_DEBUG, "Using gridmapdir. Looking for account in pool %s\n", _single_user_name);
    if (gridmapdir_userid(_gridmapdir, _dn, _single_user_name, &_pool_user_name, OVERRIDE_INCONSISTANCY) == 0){ 
      /* find_in_mapping */
      free(_single_user_name); 
      _single_user_name = NULL;
      return _pool_user_name;
    }
     
  } else {
    EEF_log(LOG_DEBUG, "Unable to find local username in gridmapfile %s\n", _gridmapfile);
    return _single_user_name;
  }

  return NULL;

}

char* find_in_mappings(const char* _dn, gridmap_list_t *_mappings){
  gridmap_line_t   *_mapping_line  = find_mapping(_mappings, _dn, MATCH_WILD_CHARS);
  gridmap_list_t   *_tmp_mapping   = NULL;
  char             *_found_mapping = NULL;
  if(_mapping_line){
    _tmp_mapping = _mappings;
    while(_tmp_mapping){
      if(_tmp_mapping->line != NULL){
        if(_tmp_mapping->line->mappings != NULL){
          _found_mapping = strdup(_tmp_mapping->line->mappings->value);
        }
      }
      _tmp_mapping= _tmp_mapping->next;
    }
  }

  return _found_mapping;
}


EES_PL_RC plugin_terminate(){
  gridmap_list_t *_mappings;
  _mappings = get_var_as_void_p("gridmappings");
  gridmap_list_free(_mappings);
  return EES_PL_SUCCESS;  
}

EES_PL_RC set_var_as_string(const char* name, char* value){
  aos_context_t*   context;
  aos_attribute_t* attribute;
  char             *attribute_name=NULL,*running_plug_name;
  int		   len;

  if(value == NULL){
    /*EEF_log(LOG_ERR, "Value cannot be NULL\n");*/
    return EES_PL_FAILURE;
  }

  rewindContexts();
  if((context = getNextContext(NONE, NULL)) == NULL){
    if ((context = createContext (NONE)) != NULL){
      addContext (context);
    } else {
      EEF_log(LOG_ERR, "Unable to get context\n");
      return EES_PL_FAILURE;
    }
  }

  /* store attribute names in AOS for lookup during run */
  if((attribute = createAttribute())){

    running_plug_name=EEF_getRunningPluginName();
    len=strlen(name)+strlen(running_plug_name)+2;
    if ((attribute_name=(char*)malloc(len))==NULL)  {
	EEF_log(LOG_ERR, "%s: out of memory\n", running_plug_name);
        return EES_PL_FAILURE;
    }
    snprintf(attribute_name, len, "%s_%s", name, running_plug_name);

    setAttributeId(attribute, attribute_name);
    free(attribute_name);
    setAttributeValue(attribute, (void*) value, strlen(value)+1);
    addAttribute(context, attribute);
  }

  return EES_PL_SUCCESS;
}


EES_PL_RC set_var_as_void_p(const char* name, void* value, size_t size){
  char             *attribute_name=NULL,*running_plug_name;
  aos_context_t*   context;
  aos_attribute_t* attribute;
  int		   len;

  /* store attribute names in AOS for lookup during run */

  rewindContexts();
  while((context = getNextContext(NONE, NULL))){
    if(context){
      break;
    }
  }   
  if(context == NULL){
    context = createContext (NONE);
    addContext (context);
  }

  if(context == NULL){
    EEF_log(LOG_ERR, "Unable to get context\n");
    return EES_PL_FAILURE;
  }

  if((attribute = createAttribute())){
    running_plug_name=EEF_getRunningPluginName();
    len=strlen(name)+strlen(running_plug_name)+2;
    if ((attribute_name=(char*)malloc(len))==NULL)  {
	EEF_log(LOG_ERR, "%s: out of memory\n", running_plug_name);
        return EES_PL_FAILURE;
    }
    snprintf(attribute_name, len, "%s_%s", name, running_plug_name);
    
    setAttributeId(   attribute, attribute_name);
    free(attribute_name);
    setAttributeValue(attribute, value, size);
    addAttribute(context, attribute);
  }
  return EES_PL_SUCCESS;
}

EES_PL_RC set_var_as_int(const char* name, int value){
  char             *attribute_name=NULL,*running_plug_name;
  aos_context_t*   context;
  aos_attribute_t* attribute;
  int		   len;

  /* store attribute names in AOS for lookup during run */

  rewindContexts();
  while((context = getNextContext(NONE, NULL))){
    if(context){
      break;
    }
  }   
  if(context == NULL){
    context = createContext (NONE);
    addContext (context);
  }

  if(context == NULL){
    EEF_log(LOG_ERR, "Unable to get context\n");
    return EES_PL_FAILURE;
  }

  if((attribute = createAttribute())){
    running_plug_name=EEF_getRunningPluginName();
    len=strlen(name)+strlen(running_plug_name)+2;
    if ((attribute_name=(char*)malloc(len))==NULL)  {
	EEF_log(LOG_ERR, "%s: out of memory\n", running_plug_name);
        return EES_PL_FAILURE;
    }
    snprintf(attribute_name, len, "%s_%s", name, running_plug_name);

    setAttributeId(   attribute, attribute_name);
    free(attribute_name);
    setAttributeValue(attribute, &value, sizeof(int));
    addAttribute(context, attribute);
  }
  return EES_PL_SUCCESS;
}

char* get_var_as_string(char* name){
  char             *localname=NULL,*running_plug_name;
  aos_context_t  * context;
  aos_attribute_t* attribute;
  int		   len;
    
  running_plug_name=EEF_getRunningPluginName();
  len=strlen(name)+strlen(running_plug_name)+2;
  if ((localname=(char*)malloc(len))==NULL)  {
      EEF_log(LOG_ERR, "%s: out of memory\n", running_plug_name);
      return NULL;
  }
  snprintf(localname, len, "%s_%s", name, running_plug_name);

  rewindContexts();
  while((context = getNextContext(NONE, NULL))){
    rewindAttributes(context);
    while((attribute = getNextAttribute(context))){
      if(strcmp(getAttributeId(attribute), localname) == 0){
        free(localname);
        return getAttributeValueAsString(attribute);
      }
    }
  }
  free(localname);
  return NULL;
}

int get_var_as_int(char* name){
  char             *localname=NULL,*running_plug_name;
  aos_context_t  * context;
  aos_attribute_t* attribute;
  int		   len;

  /* Free this */
  running_plug_name=EEF_getRunningPluginName();
  len=strlen(name)+strlen(running_plug_name)+2;
  if ((localname=(char*)malloc(len))==NULL)  {
      EEF_log(LOG_ERR, "%s: out of memory\n", running_plug_name);
      return -1;
  }
  snprintf(localname, len, "%s_%s", name, running_plug_name);

  rewindContexts();
  while((context = getNextContext(NONE, NULL))){
    rewindAttributes(context);
    while((attribute = getNextAttribute(context))){
      if(strcmp(getAttributeId(attribute), localname) == 0){
        free(localname);
        if(attribute){
          return getAttributeValueAsInt(attribute);
        }
      }
    }
  }
  free(localname);
  return -1;
}

void* get_var_as_void_p(char* name){
  char             *localname=NULL,*running_plug_name;
  aos_context_t  * context;
  aos_attribute_t* attribute;
  int		   len;

  /* Free this */
  running_plug_name=EEF_getRunningPluginName();
  len=strlen(name)+strlen(running_plug_name)+2;
  if ((localname=(char*)malloc(len))==NULL)  {
      EEF_log(LOG_ERR, "%s: out of memory\n", running_plug_name);
      return NULL;
  }
  snprintf(localname, len, "%s_%s", name, running_plug_name);

  rewindContexts();
  while((context = getNextContext(NONE, NULL))){
    rewindAttributes(context);
    while((attribute = getNextAttribute(context))){
      if(strcmp(getAttributeId(attribute), localname) == 0){
        free(localname);
        return *((void**) getAttributeValueAsVoidPointer(attribute));
      }
    }
  }
  free(localname);
  return NULL;
}

