#include <unistd.h>
#include <stdio.h>
#include <grp.h>
#include <pwd.h>
#include <getopt.h>
#include <errno.h>
#include <sys/types.h>
#include <pthread.h>
#include "eef_plugin.h"
#include "unixpriv.h"

#define UNDEFINED_ID -1
#ifdef NGROUPS_MAX
  #define NGROUPS NGROUPS_MAX
#else
  #define NGROUPS 32
#endif

static int             _maxuid;         /* max amount of uids */
static int             _maxpgid;        /* max amount of primary gids */
static int             _maxsgid;        /* max amount of secondary gids */

uid_t                  _real_uid;       /* holds real uid */
uid_t                  _saved_uid;      /* holds saved uid */

static unsigned int    _set_only_euid;  /* flag to only set effective uid */
static unsigned int    _set_only_egid;  /* flag to only set effective gid */
static unsigned int    _do_uid_check;   /* flag to check if uid exists */
static unsigned int    _dry_run;

static int             _use_uid;        /* literal uid to use */
static int             _use_gid;        /* literal gid to use */

static char            *_plugin_name;   /* name of plugin, to use when logging */

static pthread_mutex_t _mutex;          /* mutex to lock euid */

EES_PL_RC check_uid(uid_t target_uid);  /* check if uid exists on system using getpwent_r */

EES_PL_RC plugin_initialize(int argc, char* argv[]){
  static struct option long_options[] =
  {
    {"maxuid",        required_argument, 0, 'u'},
    {"maxpgid",       required_argument, 0, 'p'},
    {"maxsgid",       required_argument, 0, 's'},
    {"set_only_euid", required_argument, 0, 'e'},
    {"set_only_egid", required_argument, 0, 'g'},
    {"use_uid",       required_argument, 0, 'k'},
    {"use_gid",       required_argument, 0, 'l'},
    {"dry_run",       no_argument,       0, 'd'},
    {"check_uid",     no_argument,       0, 'c'}
  };
  int option_index, c;

  _maxuid        = UNDEFINED_ID;
  _maxpgid       = UNDEFINED_ID;
  _maxsgid       = UNDEFINED_ID;
  _real_uid      = getuid();
  _saved_uid     = geteuid();
  _use_uid       = _real_uid;
  _use_gid       = getgid();
  _dry_run       = 0;

  _set_only_euid = 0;
  _set_only_egid = 0;
  _do_uid_check  = 1;
  _plugin_name   = "posix_enf";

  EEF_log(LOG_DEBUG, "%s: Initializing posix enforcement plugin!\n", _plugin_name);

  /*_mutex = PTHREAD_MUTEX_INITIALIZER;*/

  pthread_mutex_init(&_mutex, NULL);

  opterr = 0;
  optind = 0;

  /* parse options */
  while(1){
    c = getopt_long_only(argc, argv, "u:p:s:e:g:i:dc", long_options, &option_index);
    if(c == -1){
      break;
    }
    switch(c){
      case 'u':
        if((_maxuid = strtol(optarg, NULL, 10)) == UNDEFINED_ID){
          EEF_log(LOG_ERR, "%s: %s", _plugin_name, strerror(errno)); 
        }
        break;
      case 'p':
        if((_maxpgid = strtol(optarg, NULL, 10)) == UNDEFINED_ID){
          EEF_log(LOG_ERR, "%s: %s", _plugin_name, strerror(errno)); 
        }
        break;
      case 's':
        if((_maxsgid = strtol(optarg, NULL, 10)) == UNDEFINED_ID){
          EEF_log(LOG_ERR, "%s: %s", _plugin_name, strerror(errno)); 
        }
        break;
      case 'e':
        if(strcmp(optarg,"yes") == 0){
          _set_only_euid = 1;
        }
        break;
      case 'g':
        if(strcmp(optarg, "yes") == 0){
          _set_only_egid = 1;
        }
        break;
      case 'k':
        if((_use_uid = strtol(optarg, NULL, 10)) == UNDEFINED_ID){
          EEF_log(LOG_ERR, "%s: %s", _plugin_name, strerror(errno)); 
        }
        break;
      case 'l':
        if((_use_gid = strtol(optarg, NULL, 10)) == UNDEFINED_ID){
          EEF_log(LOG_ERR, "%s: %s", _plugin_name, strerror(errno)); 
        }
        break;
      case 'c':
        _do_uid_check = 1;
        break;
      case 'd':
        _dry_run = 1;
        break;
      case '?':
        EEF_log(LOG_ERR, "%s: unknown option %s", _plugin_name, optarg);
        break;
      case ':':
        EEF_log(LOG_ERR, "%s: missing argument for %s", _plugin_name, optarg);
        break;
    }
  }

  /* sanity checks */
  if(_maxsgid > NGROUPS){
    EEF_log(LOG_ERR, "%s: Option -maxsgid %i exceeds the system limit of %i", _plugin_name, _maxsgid, NGROUPS);
    return EES_PL_FAILURE;
  } else {
    if(_maxsgid == UNDEFINED_ID){
      _maxsgid = NGROUPS;
      EEF_log(LOG_NOTICE, "%s: Option -maxsgid defaulted to maximum %i", _plugin_name, NGROUPS);
    }
  }

  EEF_log(LOG_INFO, "%s: Initialized plugin posix_enf with options:\n", _plugin_name);
  EEF_log(LOG_INFO, "%s: _maxuid:        %i\n", _plugin_name, _maxuid);
  EEF_log(LOG_INFO, "%s: _maxpgid:       %i\n", _plugin_name, _maxpgid);
  EEF_log(LOG_INFO, "%s: _maxsgid:       %i\n", _plugin_name, _maxsgid);

  EEF_log(LOG_INFO, "%s: _set_only_euid: %i\n", _plugin_name, _set_only_euid);
  EEF_log(LOG_INFO, "%s: _set_only_egid: %i\n", _plugin_name, _set_only_egid);

  EEF_log(LOG_INFO, "%s: _use_uid:       %i\n", _plugin_name, _use_uid);
  EEF_log(LOG_INFO, "%s: _use_gid:       %i\n", _plugin_name, _use_gid);
  EEF_log(LOG_INFO, "%s: _check_uid:     %i\n", _plugin_name, _do_uid_check);

  /* save effective uid and switch to real uid */
  downgradeEffectiveToRealUid(&_real_uid, &_saved_uid);

  return EES_PL_SUCCESS;
}

EES_PL_RC plugin_run(){
  uid_t            _target_uid;
  gid_t            _target_gid;
  /*sudo_fork_t      _fork_struct;*/

  aos_context_t    *_context        = NULL;
  aos_attribute_t  *_attribute      = NULL;

  /* initialize _target_uid and _target_gid with values associated with user 'nobody' */
  _target_uid = threadsafe_getuid_from_name("nobody");
  _target_gid = threadsafe_getgid_from_name("nogroup");

  /* if an explicit uid was not passed, check the AOS */
  if(_use_uid == -1){
    rewindContexts();
    while((_context = getNextContext(OBLIGATION, NULL))){
      if(strcmp(getContextObligationId(_context), "uidgid") == 0){
        rewindAttributes(_context);
        while((_attribute = getNextAttribute(_context))){
          if(strcmp(getAttributeId(_attribute), "posix-uid") == 0){
            _target_uid = getAttributeValueAsInt(_attribute);
          }
        }
      }
    }
  } else {
    /* otherwise, use the supplied uid */
    _target_uid = _use_uid;
  }

  /* if an explicit gid was not passed, check the AOS */
  if(_use_gid == -1){
    rewindContexts();
    while((_context = getNextContext(OBLIGATION, NULL))){
      if(strcmp(getContextObligationId(_context), "uidgid") == 0){
        rewindAttributes(_context);
        while((_attribute = getNextAttribute(_context))){
          if(strcmp(getAttributeId(_attribute), "posix-gid") == 0){
            _target_gid = getAttributeValueAsInt(_attribute);
          }
        }
      }
    }
  } else {
    /* otherwise, use the supplied uid */
    _target_gid = _use_gid;
  }

  if(_do_uid_check){
    if(check_uid(_target_uid) != EES_PL_SUCCESS){
      return EES_PL_FAILURE;
    }
  }

  /*if((_fork_struct = calloc(1, sizeof(sudo_fork_t))) == NULL){*/
  /*return EES_PL_FAILURE;*/
  /*}*/

  /* switch to saved uid in order to gain user-switching privileges */
  if(!_dry_run){
    upgradeEffectiveToRealUid(_real_uid, _saved_uid);

    if(_set_only_egid){
      /* only set effective gid */
      /*_fork_struct->egid = _target_gid;*/
      if(setegid(_target_gid) != 0){
        goto fail_group;
      }
      /* check if we haven't switched to root group */
      if(getegid() == 0){
        goto fail_group;
      }
    } else {
      /* set both real and effective gid */
      if(setregid(_target_gid, _target_gid) != 0){
        goto fail_group;
      }
      /* check if we haven't switched to root group */
      if(getgid() == 0 && getegid() == 0){
        goto fail_group;
      }
    }

    if(_set_only_euid){
      /* only set effective uid */
      if(seteuid(_target_uid) != 0){
        goto fail_user;
      }
      if(geteuid() == 0){
        /* check if we haven't switched to root */
        goto fail_user;
      }
    } else {
      /* set both real and effective gid */
      if(setreuid(_target_uid, _target_uid) != 0){
        goto fail_user;
      }
      if(getuid() == 0 && geteuid() == 0){
        /* check if we haven't switched to root */
        goto fail_user;
      }
    }

  }
  /*pthread_mutex_unlock(&_mutex);*/

  return EES_PL_SUCCESS;

/* may not switch to group associated with root */
fail_group:
  EEF_log(LOG_ERR, "%s: unable to set group id to: %i. %s\n", _plugin_name, _target_gid, strerror(errno));
  goto fail_enforcement;
fail_user:
  EEF_log(LOG_ERR, "%s: unable to set user id to: %i. %s\n", _plugin_name, _target_uid, strerror(errno));
  goto fail_enforcement;

fail_enforcement:
  return EES_PL_FAILURE;
}

/* terminate plugin */
EES_PL_RC plugin_terminate(){
  EEF_log(LOG_INFO, "%s: plugin posix_enf terminated\n", _plugin_name);
  return EES_PL_SUCCESS;
}

EES_PL_RC check_uid(uid_t target_uid){
  struct  passwd   _pw_entry;
  struct  passwd   *_pw_entry_p     = &_pw_entry;
  struct  passwd   *_tmp_pw_entry_p = NULL;
  char             _pw_buffer[200];
  size_t           _pw_size         = sizeof(_pw_buffer);

  EEF_log(LOG_DEBUG, "%s: Checking uid %i\n", _plugin_name, target_uid);
  errno = getpwuid_r(target_uid, _pw_entry_p, _pw_buffer,  _pw_size, &_tmp_pw_entry_p);
  if(_tmp_pw_entry_p == NULL){
    if(errno == 0){
      EEF_log(LOG_ERR, "%s: No user with uid '%s' found", _plugin_name, target_uid); 
    } else {
      EEF_log(LOG_ERR, "%s: %s", _plugin_name, strerror(errno)); 
    }
    endpwent();
    return EES_PL_FAILURE;
  }
  EEF_log(LOG_DEBUG, "User name %s\n", _pw_entry.pw_name );
  EEF_log(LOG_DEBUG, "Uid %i\n", _pw_entry.pw_uid );
  EEF_log(LOG_DEBUG, "Gid %i\n", _pw_entry.pw_gid );
  EEF_log(LOG_DEBUG, "Initial dir %s\n", _pw_entry.pw_dir );
  EEF_log(LOG_DEBUG, "Shell %s\n", _pw_entry.pw_shell );

  endpwent();

  return EES_PL_SUCCESS;

}

