#include "plugin_manager.h"
#include "_aos.h"
#include "eef_aos.h"

/* Thread-local storage space for AOS data */
static pthread_key_t           _aos_key;
/* Normally not okay to do initialisation outside of a function, but this way it (hopefully) will only be set one - during compile time */
static pthread_once_t  	       _aos_key_once = PTHREAD_ONCE_INIT;
/* The backup storage state, used to store references to aos_context_t instances globally, used when no thread-local attributes were found */
static struct aos_state_s *    _global_state;

/* Flags for runtime checks */
static int                     _is_initialized;
static int                     _is_threading;

/* internal methods */

/**
 *  returns _is_initialized
 */
int aos_is_initialized(){
	return _is_initialized;
}

/**
 * sets threading flag
 */
void aos_start_threading(){
	_is_threading = 1;
}

/**
 * returns a pointer to an aos_storage_t struct that corresponds to the current thread context. If the current context is not threaded, a pointer to a global storage struct is returned.
 */
aos_state_t * aos_get_state(void){
  pid_t             tid;
  aos_state_t *     state = NULL;

  /* If we are not in threading mode, no thread-local storage should be accessed */
  if(!_is_threading){
    return _global_state;
  }

  /* Get current thread id */
  tid = syscall(SYS_gettid);
  /*tid = gettid();*/

  /* Try to create thread-local storage handle if it does not exist yet for this thread, return the existing aos_state_t otherwise */
  /* TODO generalize with AOS_Init */
  state = pthread_getspecific(_aos_key);
  if(state == 0) {
    /* Allocate and initialize a new aos_state_t struct */
    if((state = calloc(1, sizeof(aos_state_t)))){
      state->last_context = NULL;
      /* Initialize the aos_storage_t linked list in the thread-local storage field */
      if((state->saved_storage = calloc(1, sizeof(aos_storage_t)))){
        /* Set the head node to NULL */
        state->saved_storage->list_contexts = NULL;
        /* Set this as the current storage  */
        state->current_storage = state->saved_storage;
        /* Set the handle in the pthread thread-local storage */
        pthread_setspecific(_aos_key, state);
        return state;
      }
    }
  } else {
    return state;
  }
  return NULL;
}

/**
 * Return the current storage from the current state - thread-local or global storage
 */
aos_storage_t* aos_get_storage(){
  aos_state_t* current_state = aos_get_state();
  return current_state->current_storage;
}

/**
 * Wrapper to clean up an aos_state_t by calling aos_free_storage, NULLing fields and freeing the struct
 */
EES_RC aos_free_state(aos_state_t* state){
  aos_free_storage(state->saved_storage);
  state->saved_storage = NULL;
  state->last_context  = NULL;
  free(state);
  return EES_SUCCESS;
}

/**
 * Frees all members of aos_storage_t structs
 */
EES_RC aos_free_storage(aos_storage_t* storage){
  /* Temporary pointers to aos_context_t and aos_attribute_t */
  aos_context_t   *context        = NULL, *next_context    = NULL;
  aos_attribute_t *attribute      = NULL, *next_attribute  = NULL;
  /* The number of attributes we've found */
  int             attribute_count = 0;

  /* We need a valid pointer to a storage object to iterate the linked list of aos_context_t structs */
  if(!storage){
    return EES_FAILURE;
  }

  EEF_log(LOG_DEBUG, "Cleaning aos storage at %p\n", storage);

  /* Get the head node of the list of contexts*/
  context = storage->list_contexts;

  /* Walk the list */
  while(context){
    /* Get the list of attributes from the aos_context_t */
    attribute = context->list_attributes;
    /* Walk this list too */
    while(attribute){
      /* Free heap data fields */
      free(attribute->id);
      free(attribute->issuer);
      /*free(attribute->type);*/

      /* Some fields may point to automaticly allocated variables */
      if(attribute->needs_free){
        EEF_log(LOG_DEBUG, "Freeing arg data %i at %p\n",attribute_count, attribute->data);
        free(attribute->data);
      }

      /* Go to the next attribute in the list, and free the current one */
      next_attribute = attribute->next;
      free(attribute);
      attribute = next_attribute;
      attribute_count++;
    }

    EEF_log(LOG_DEBUG, "Freeing context at %p\n", context);
    /* Go to the next context in the list, and free the current one */
    next_context = context->next;
    free(context);
    context = next_context;
  }

  /* Free the storage node that contained the contexts */
  free(storage);
  storage = NULL;
  return EES_SUCCESS;
}

/* base functions */

/**
 * Returns a pointer to a freshly allocated aos_context_t struct
 */
aos_context_t * createContext(aos_context_class_t context_class){
  aos_context_t * context;
  if((context = calloc(1, sizeof(aos_context_t)))){
    context->context_class = context_class;
    context->list_attributes = NULL;
    context->last_attribute = NULL;
    context->next = NULL;
  }
  return context;
}

/**
 * Adds supplied context to storage attained through aos_get_storage()
 */
EES_RC addContext(aos_context_t* context){
  /* Get the current - thread-local or global - storage */
  aos_state_t*     state = aos_get_state();
  aos_storage_t* storage = state->saved_storage;
  /* Temporary aos_context_t pointer */
  aos_context_t* tmp_context = NULL;

  if(!storage){
    EEF_log(LOG_ERR, "No storage available at %p", storage);
    return EES_FAILURE;
  }

  EEF_log(LOG_DEBUG, "Adding context at %p to storage at %p\n", context, storage);
  if(storage->list_contexts != NULL){
    /* We need to walk the list before appending to it */
    if((tmp_context = storage->list_contexts) != NULL){
      while(tmp_context->next){
        /* Duplicate check */
        if(tmp_context == context){
          /*EEF_log(LOG_ERR, "The context you're trying to add was already added!\n");*/
          return EES_FAILURE;
        }
        tmp_context = tmp_context->next;
      }
    }
    /* Add the context to the end of the list */
    tmp_context->next = context;
    EEF_log(LOG_DEBUG, "Inserted at %p\n", tmp_context);
  } else {
    /* We are creating the head node */
    storage->list_contexts = context;
    EEF_log(LOG_DEBUG, "Created at %p\n", context);
  }
  
  /* Set the pointer used by getNextContext to the current context */
  aos_set_iterator(context);

  /*eef_log(LOG_DEBUG, "Copied context to %p\n", storage->last_context);*/
  return EES_SUCCESS;
}

/* Set the iterator for the current storage to the thread-local storage, and sets the aos_context_t* iterator to the head node*/
/* NOTE not using storage parameter anymore */
EES_RC rewindContexts(void){
  /* Get the current aos_state_t - either thread-local storage or _global_state */
  aos_state_t*    state = aos_get_state();

  if(state){
    state->current_storage = state->saved_storage;
    if(state->current_storage){
      /* Set the iterator for the list of aos_context_t to the head node */
      aos_set_iterator(state->current_storage->list_contexts);
      return EES_SUCCESS;
    } else {
      EEF_log(LOG_ERR, "Unable to aquire thread-local storage handle in AOS::rewindContexts()\n");
    }
  }
  return EES_FAILURE;
}

/**
 * Used to iterate the list of contexts from the current state
 */
EES_RC aos_set_iterator(aos_context_t* context){
  /* Get the current aos_state_t - either thread-local storage or _global_state */
  aos_state_t* state = aos_get_state();
  if(state->current_storage){
    /* Set the lookup variable to the provided context */
    state->last_context = context;
    return EES_SUCCESS;
  } else {
    EEF_log(LOG_ERR, "Invalid storage passed to aos_set_iterator(): %p\n", state->current_storage);
  }
  return EES_FAILURE;
}

aos_context_t* aos_get_iterator(void){
  /* Get the current aos_state_t - either thread-local storage or _global_state */
  aos_state_t* state = aos_get_state();
  if(state->current_storage){
    /* Return the lookup variable set by getNextContext or aos_set_iterator */
    if(state->last_context){
      return state->last_context;
    }
  } else {
    EEF_log(LOG_ERR, "Invalid storage passed to aos_set_iterator(): %p\n", state->current_storage);
  }
  return NULL;
}

/**
 * returns the next aos_context_t pointer with the supplied context_class from the list of contexts in the supplied storage.
 * will fall through and try to use global storage in case nothing was found
 */
aos_context_t * getNextContext(aos_context_class_t context_class, aos_storage_t* storage){
  /* Get the current aos_state_t - either thread-local storage or _global_state */
  aos_state_t*    state           = aos_get_state();
  /* Get the current context */
  aos_context_t*  current_context = aos_get_iterator();

  /* We need to find a specific context type */
  if(context_class != ANY){
    while(current_context != NULL){
      if((current_context->context_class == context_class)){
        /* Found context */
        break;
      }
      current_context = current_context->next;
    }
  }

  /* If no context was found, */
  if((current_context == NULL) && 
        /* and we are not requested to be using the global storage already, */
        (storage != _global_state->current_storage) &&
        /* and the current storage in the local or global state is also not the global storage */
        (state->current_storage != _global_state->current_storage)){

    /* Set the current state's current storage to the global storage */
    state->current_storage = _global_state->current_storage;
    /* Set the current state iterator to the head node of the list of contexts */
    aos_set_iterator(_global_state->current_storage->list_contexts);
    /* Recurse */
    current_context = getNextContext(context_class, _global_state->current_storage);
  }

  /* Save iterator */
  if(current_context){
    aos_set_iterator(current_context->next);
  } else {
    aos_set_iterator(NULL);
  }

  return current_context;
}

/**
 * Sets the obligation name in the aos_context_t pointed to by context to obligation
 */
void setContextObligationId(aos_context_t* context, char * obligation){
  if(context->context_class == OBLIGATION){
    context->obligation_name = obligation;
  }
}

/**
 * Returns the obligation name from the aos_context_t pointed to by context
 */
char* getContextObligationId(aos_context_t* context){
  return context->obligation_name;
}

/**
 * Returns the context class from the aos_context_t pointed to by context
 */
aos_context_class_t getContextClass(aos_context_t* context){
  return context->context_class;
}

/**
 * Returns a pointer to a freshly allocated aos_attribute_t struct
 */
aos_attribute_t* createAttribute(void){
  aos_attribute_t* attribute = NULL;
  if((attribute = calloc(1, sizeof(aos_attribute_t)))){
    attribute->next = NULL;
    attribute->data = NULL;
    attribute->setting_plugin = get_running_plugin();
  }
  return attribute;
}

/**
 * Adds aos_attribute_t pointed to by attribute to the list of attributes in aos_context_t pointed to by context
 */
EES_RC addAttribute(aos_context_t * context, aos_attribute_t* attribute){
  aos_attribute_t* working = NULL;
  if (context->list_attributes == NULL){
    context->list_attributes = attribute;
  } else {
    working = context->list_attributes;
    while(working->next)
      working = working->next;
    working->next = attribute;
  }
  return EES_SUCCESS;
}

EES_RC rewindAttributes(aos_context_t* context){
  context->last_attribute = context->list_attributes;
  return EES_SUCCESS;
}

/**
 * Returns the next aos_attribute_t pointer from list of attributes in the supplied context
 */
aos_attribute_t* getNextAttribute(aos_context_t* context){
  aos_attribute_t* attribute = NULL;
  if(!context){
    return NULL;
  }

  if(context->last_attribute){
    attribute = context->last_attribute;
  }
  if(attribute){
    context->last_attribute = attribute->next;
  } else {
    context->last_attribute = attribute;
  }
  return attribute;
}

/**
 * Removes a supplied attribute from the supplied context
 */
EES_RC destroyAttribute(aos_context_t* context, aos_attribute_t* attribute){
  aos_attribute_t *last_attribute = NULL, *current_attribute = NULL;
  for(current_attribute = context->list_attributes; current_attribute != NULL; last_attribute = current_attribute, current_attribute = current_attribute->next){
    if(current_attribute == attribute){
      if(current_attribute->setting_plugin != get_running_plugin()){
        EEF_log(LOG_ERR, "Argument %s is not owned by running plugin %s", current_attribute->id, get_running_plugin()->name);
        return EES_FAILURE;
      }
      free(attribute->data);
      free(attribute);
      attribute = NULL;
      if(last_attribute == NULL){
        context->list_attributes = current_attribute->next;
      } else {
        last_attribute->next = current_attribute->next;
      }
      break;
    }
  }

  if(current_attribute){
    free(current_attribute->data);
    free(current_attribute);
    current_attribute = NULL;
    return EES_SUCCESS;
  }
  return EES_FAILURE;
}

/**
 * Sets attribute id in supplied attribute
 */
EES_RC setAttributeId(aos_attribute_t* attribute, const char* id){
  free(attribute->id);
  attribute->id = NULL;
  if(id != NULL){
    attribute->id = strdup(id);
  } else {
    EEF_log(LOG_ERR, "setAttributeId(), tried to set NULL id\n");
  }
  return EES_SUCCESS;
}

/**
 * Sets attribute issuer in supplied attribute
 */
EES_RC setAttributeIssuer(aos_attribute_t* attribute, const char* issuer){
  free(attribute->issuer);
  attribute->issuer = NULL;
  if(issuer != NULL){
    attribute->issuer = strdup(issuer);
  /*} else {*/
    /*EEF_log(LOG_ERR, "setAttributeIssuer(), tried to set NULL issuer\n");*/
    /*return EES_FAILURE;*/
  }
  return EES_SUCCESS;
}

/**
 * Sets attribute value in supplied attribute
 */
EES_RC setAttributeValue(aos_attribute_t* attribute, const void* value, size_t size){
  if(value != NULL){
    if(attribute->needs_free == 1){
      free(attribute->data);
      attribute->data = NULL;
      attribute->needs_free = 0;
    }
    if(size == 0) {
      attribute->data = value;
      attribute->needs_free = 0;
    } else {
      attribute->data = calloc(1, size+1);
      memcpy(attribute->data, value, size);
      attribute->needs_free = 1;
    }
  } else {
    return EES_FAILURE;
  }
  return EES_SUCCESS;
}

/**
 * Sets attribute type in supplied attribute
 */
EES_RC setAttributeType(aos_attribute_t * attribute, const char* type){
  free(attribute->type);
  attribute->type = NULL;
  if(type != NULL)
      attribute->type = strdup(type);
  return EES_SUCCESS;
}

/**
 * Returns type from supplied attribute
 */
char* getAttributeType(aos_attribute_t * attribute){
  return attribute->type;
}

/**
 * Returns id from supplied attribute
 */
char* getAttributeId(aos_attribute_t* attribute){
  return attribute->id;
}

/**
 * Returns issuer from supplied attribute
 */
char* getAttributeIssuer(aos_attribute_t* attribute){
  return attribute->issuer;
}

/**
 * Returns data from supplied attribute casted to a char*
 */
char* getAttributeValueAsString(aos_attribute_t* attribute){
  return (char*) attribute->data;
}

/**
 * Returns data from supplied attribute casted to a int
 */
int getAttributeValueAsInt(aos_attribute_t* attribute){
  return *((int*) attribute->data);
}

/**
 * Returns data from supplied attribute casted to a void*
 */
void* getAttributeValueAsVoidPointer(aos_attribute_t* attribute){
  return (void*) attribute->data;
}

/* Main AOS control functions */

/**
 * Initializes the AOS and creates thread-local storage key
 */
EES_RC AOS_Init (void){
  _is_threading = 0;
  pthread_once(&_aos_key_once, aos_make_key);

  /* TODO generalize with aos_get_state() */
  /* Allocate and initialize a new aos_state_t struct */
  if((_global_state = calloc(1, sizeof(aos_state_t)))){
    /* Set the last_context lookup field to NULL */
    _global_state->last_context = NULL;
    /* Initialize the aos_storage_t linked list in the thread-local storage field */
    if((_global_state->saved_storage = calloc(1, sizeof(aos_storage_t)))){
      /* Set the head node to NULL */
      _global_state->saved_storage->list_contexts = NULL;
      /* Set this as the current storage  */
      _global_state->current_storage = _global_state->saved_storage;
      /* The AOS is initialized */
      _is_initialized = 1;
      return EES_SUCCESS;
    }
  }

  return EES_FAILURE;
}

/**
 * Creates thread-local storage key
 * This is a callback for pthread_once
 */
void aos_make_key(void){
  pthread_key_create(&_aos_key, aos_free_key);
}

/**
 * Frees thread-local storage key - callback for pthread_key_create
 */
void aos_free_key(void* state){
  pid_t             tid;
  aos_state_t*      tmp_state = (aos_state_t*) state;
  tid = syscall(SYS_gettid);
  EEF_log(LOG_DEBUG, "Freeing thread-local storage for thread %i\n", tid);
  aos_free_state(tmp_state);
}

/**
 * Frees storage for this state - only really applies for thread-local storage
 */
EES_RC AOS_Clean(void){
  pid_t             tid;
  if(_is_threading){
    tid = syscall(SYS_gettid);
    EEF_log(LOG_DEBUG, "Killing thread %i\n", tid);
    aos_free_state(aos_get_state());
  }

  /* _global_state will be freed by AOS_Term */
  return EES_SUCCESS;
}

/**
 * Terminates AOS, deleting the global storage key
 */
EES_RC AOS_Term (void){
  _is_threading = 0;
  /* Free the global aos_state_t */
  aos_free_state(_global_state);

  /* Remove thread-local storage key */
  pthread_setspecific(_aos_key, NULL);
  pthread_key_delete(_aos_key);

  /* AOS terminated */
  _is_initialized = 0;
  return EES_SUCCESS;
}

/**
 * Dumps out all the data present in the AOS, as seen from the current thread's perspective
 */
EES_RC aos_dump_argslist (void) {
  /* Temporary aos_context_t and aos_attribute_t pointers */
  aos_context_t   *context         = NULL;
  aos_attribute_t *attribute       = NULL;
  /* Temporary char*s to hold names and values */
  char            *attribute_name  = NULL;
  char            *attribute_value = NULL;

  EEF_log(LOG_DEBUG, "Dumping aos");

  rewindContexts();

  while((context = getNextContext(ANY, NULL)) != NULL){
    EEF_log(LOG_DEBUG, "Context %p class: %i\n", context, context->context_class);
    rewindAttributes(context);
    while((attribute = getNextAttribute(context)) != NULL){
      attribute_name = getAttributeId(attribute);
      attribute_value = getAttributeValueAsString(attribute);
      if(attribute_name && attribute_value){
        if(attribute->needs_free){
          EEF_log(LOG_DEBUG, "\t%s=%s\n", attribute_name, attribute_value);
        } else {
          EEF_log(LOG_DEBUG, "\t%s=%i\n", attribute_name, attribute_value);
        }
      }
    }
  }

  return EES_SUCCESS;
}

