#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

#include <unistd.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <string.h>
#include <netdb.h>
#include <arpa/inet.h>

#include "ees_eics.h"
#include "config.h"
#include "xacml.h"
#include "xacml_server.h"
#include "eef_library.h"
#include "cgul/log_to_file/log_to_file.h"
#include "cgul_patched/network/net_common.h"

#define _MULTI_THREADED
#define BACKLOG 100

char* config_file_path;                 /* config file path */
xacml_server_t server;                  /* xacml server object */
int signal_is_being_handled = 0;        /* signal handler state/lock */

void signal_handler(int sig);           /* main signal handler */
EES_RC process_xacml(int port);         /* creates and starts new xacml server instance */
EES_RC ees_start(void);                 /* initializes EEF */
EES_RC ees_stop(void);                  /* terminates EEF */
EES_RC ees_loop(int port);              /* main loop - calls process_xacml and pauses */

/* callback for libxacml - thread body */
int ees_xacml_authorize(void *handler_arg,
      const xacml_request_t request,
      xacml_response_t response);

unsigned int running_threads;
pthread_mutex_t stopping;
int running;
pthread_mutex_t thread_count;
pthread_cond_t idle;

/**
 * Prepares logging, initializes EEF and puts the EEF in threading mode
 */
EES_RC ees_start(void){
  const char* logfile=NULL;
  /* set log options */
  cgul_set_ident(BINARY_NAME);
  /* CGUL_STDERR is only used for errors and warnings */
  /*cgul_set_log_options(CGUL_SYSLOG | CGUL_STDERR);*/
  cgul_set_log_options(CGUL_SYSLOG);

  /* reuse this part when implementing SIGHUP */
/*  cgul_log(LOG_DAEMON, "Parsing config file %s\n", config_file_path);*/
  cgul_log(LOG_DEBUG, "Parsing config file %s\n", config_file_path);
  if(EEF_init(config_file_path, cgul_log) == EES_FAILURE){
    /* at this point, EEF_log can/will be a reference to the second argument passed to EEF_init */
    cgul_log(LOG_ERR, "Failed to parse config file (%s), exiting...", config_file_path);
    return EES_FAILURE;
  }
  /* Now we should have a logfile or not... */
  logfile = EEF_get_log_file();

  EEF_log(LOG_DEBUG, "Done parsing. Log file at: %s\n", logfile);
  /* Try to log to the logfile specified in the pdl configuration */
  if(logfile){
    cgul_set_log_options( (cgul_get_log_options() | CGUL_LOGFILE) & ~CGUL_SYSLOG);
    if(cgul_openlog(logfile) != 0){
      EEF_log(LOG_ERR, "Unable to open log file: %s\n", logfile);
      return EES_FAILURE;
    }
  }

  /* Set threading flag in AOS and plugin manager */
  EEF_startThreading();

  pthread_mutex_lock(&thread_count);
  running_threads = 0;
  running = 1;
  pthread_mutex_unlock(&thread_count);
  EEF_log(LOG_NOTICE, "EES Ready.\n");

  return EES_SUCCESS;
}

/**
 * Main EES loop. Starts xacml server and pauses.
 */
EES_RC ees_loop(int port){
  while(1){ /* this is to prevent falling out of the loop when a signal gets handled */
    signal_is_being_handled = 0;
    if(process_xacml(port) == EES_SUCCESS){
      /*printf("Paused.\n");*/
      pause();
    } else {
      EEF_log(LOG_ERR, "Unable to start xacml server\n");
      ees_stop();
      exit(EXIT_FAILURE);
    }
  }
  return EES_SUCCESS;
}

/**
 * Terminates EES
 */
EES_RC ees_stop(void){
  /* Wait until the service is idle */
  pthread_mutex_lock(&thread_count);
  running = 0;
  if(running_threads > 0){
    EEF_log(LOG_DEBUG, "Waiting for %i running threads\n", running_threads);
    pthread_cond_wait(&idle, &thread_count);
  }
  pthread_mutex_unlock(&thread_count);

  /* Stop xacml server */
  xacml_server_destroy(server);
  EEF_log(LOG_NOTICE, "Stopped XACML server\n");
  /* Clean thread-local AOS storage */
  AOS_Clean();
  EEF_log(LOG_NOTICE, "Cleaned AOS\n");
  /* Stop EEF library - evaluation manager, plugin manager and the AOS */
  if(EEF_term() == EES_SUCCESS){
    EEF_log(LOG_NOTICE, "EEF terminated\n");
  } else {
    EEF_log(LOG_NOTICE, "Error while terminating EEF\n");
  }
  fflush(stdout);
  fflush(stderr);
  EEF_log(LOG_NOTICE, "EES Stopped\n");  
  cgul_closelog();
  return EES_SUCCESS;
}

/**
 * Creates and starts xacml server.
 */
EES_RC process_xacml(int port){
  int fd;
  /*shutdown(fd, 0);*/
  if((fd = createAndSetUpATCPServerSocket(port, BACKLOG)) > -1){
    xacml_init(); /* if you look at the code, you'll see this function is empty */
    if(xacml_server_init(&server, ees_xacml_authorize, NULL) == XACML_RESULT_SUCCESS){
      xacml_server_set_fd(server, fd);
      xacml_server_start(server);
      return EES_SUCCESS;
    }
  }
  return EES_FAILURE;
}

int ees_xacml_authorize(void *handler_arg, const xacml_request_t request, xacml_response_t response) {
  pid_t             tid;

  tid = syscall(SYS_gettid);
  EEF_log(LOG_DEBUG, "In thread %i\n", tid);

  pthread_mutex_lock(&thread_count);
  if(!running){
    goto end;
  }
  running_threads++;
  pthread_mutex_unlock(&thread_count);

  xacml_response_set_issuer(response, "EES");

  extractRequestSubjectToAOS (request);
  extractRequestActionToAOS (request);
  extractRequestResourceToAOS (request);
  extractRequestEnvironmentToAOS (request);

  if(EEF_run() == EES_FAILURE) { 
    constructFailureResponse(&response);
    EEF_log(LOG_ERR, "Failed to run the EEF successfully\n");
  } else {
    constructResponse(&response);
  }

  pthread_mutex_lock(&thread_count);
  running_threads--;

end:
  if(running == 0 && running_threads == 0){
    pthread_cond_signal(&idle);
  }
  pthread_mutex_unlock(&thread_count);

  return 0;
}

/**
 * Handles incoming signals
 */
void signal_handler(int sig){
  pid_t             tid;
  tid = syscall(SYS_gettid);
  int port;
  EEF_log(LOG_DEBUG, "Got signal %i\n", sig);
  if(signal_is_being_handled == 0){
    signal_is_being_handled = 1;
    switch(sig){
      case SIGHUP:
        EEF_log(LOG_NOTICE, "Got SIGHUP - restart \n");
        ees_stop();
        if(cgul_openlog(EEF_get_log_file()) == 0){
          if(ees_start() == EES_SUCCESS){
#if !ENABLE_DEBUG
            EEF_daemonize();
#endif
            if((port=EEF_get_port()) == -1) {
              EEF_log(LOG_ERR, "Port is not set or invalid.\n");
              exit(EXIT_FAILURE);
            }
	    /* Now we are the new process, since the parent has quit in the
	     * EEF_daemonize() */
            return;
          }
        }
        fprintf(stderr, "Unable to restart - check logs.\n");
        exit(EXIT_FAILURE);
        break;
      case SIGTERM:
        EEF_log(LOG_NOTICE, "Got SIGTERM - shutting down\n");
        ees_stop();
        exit(EXIT_SUCCESS);
        break;
      case SIGUSR1:
        cgul_closelog();
        if(cgul_openlog(EEF_get_log_file()) != 0){
          EEF_log(LOG_ERR, "Unable to open log file: %s\n", EEF_get_log_file());
          exit(EXIT_FAILURE);
        }
        break;
      default:
        break;
    }
  }
}

/**
 * Main
 */
int main (int argc, char* argv[]){
  struct sigaction sa;
  int port;

  if(argc > 1){
    config_file_path = argv[1];
  } else {
    config_file_path = EES_CONFIG_FILE;
  }

  if(!config_file_path){
    fprintf(stderr, "Unable to open config file %s\n", config_file_path);
  }
  cgul_log(LOG_DEBUG, "Opening config file %s\n", config_file_path);

  if(sigemptyset(&sa.sa_mask)){
    cgul_log(LOG_ERR, "Error initializing signal handlers\n");
  }

  /* catch hangup and term signals */

  sa.sa_handler = signal_handler;
  sa.sa_flags = SA_SIGINFO | SA_RESTART;

  if(sigaction(SIGHUP, &sa, NULL)){
    cgul_log(LOG_ERR, "Unable to add signal handler for SIGHUP\n");
  }
  if(sigaction(SIGTERM, &sa, NULL)){
    cgul_log(LOG_ERR, "Unable to add signal handler for SIGHUP\n");
  }
  /*if(sigaction(SIGUSR1, &sa, NULL)){*/
  /*cgul_log(LOG_ERR, "Unable to add signal handler for SIGUSR1\n");*/
  /*}*/

#if ENABLE_DEBUG
  cgul_set_log_options(cgul_get_log_options() | CGUL_STDDEBUG);
#endif

  if(ees_start() == EES_SUCCESS){
#if !ENABLE_DEBUG
    EEF_daemonize();
#endif
    /*printf("port=%d\n",EEF_get_port());*/
    if((port=EEF_get_port()) == -1) {
      EEF_log(LOG_ERR, "Port is not set or invalid.\n");
      return EXIT_FAILURE;
    }

    ees_loop(port);
    return EXIT_SUCCESS;
  }
  fprintf(stderr, "Unable to start - check logs.\n");

  return EXIT_FAILURE;

}

