#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fnmatch.h>
#include <unistd.h>

#include "gridmapfile.h"
#include "fileutil.h"

#define WHITESPACE_CHARS        " \t\n"
#define QUOTING_CHARS           "\""
#define ESCAPING_CHARS          "\\"
#define COMMENT_CHARS           "#"
#define WILD_CHARS              '*'
#define DIGIT_CHARS             "0123456789"
#define USERID_SEP_CHARS        ","
#define USERID_TERMINATOR_CHARS USERID_SEP_CHARS WHITESPACE_CHARS
#define NUL                     '\0'
#define USERID_CHUNK_SIZE       4


/******************************************************************************
Function:   gridmap_parse_line
Description:
    Given a line from the gridmap file, parse it returning
    a gridmap_line_t structure. line is modified during parsing.

    The format of the line is expected to be:

    <DN> <userid>[,<userid>[,<userid>...]]

    Leading and trailing whitespace is ignored.

    userids must only have a comma between them, no whitespace.

    Anything after the userids is ignored.

    Anything after an unescaped comment character is ignored.

Parameters:
    line, a pointer to the line from the file (NUL-terminated string)

    gline, a pointer to a pointer that will be set to point at
    the gridmap_line_t structure containing the line information.
    If the line contains no content, gline will be set to NULL.

Returns:
    0 on success, non-zero on error.

******************************************************************************/
static int gridmap_parse_line(char * line, gridmap_line_t ** gline);
static int xdigit_to_value(char xdigit);
static int gridmap_parse_globusid(const char * unparsed, char ** pparsed);
static void gridmap_free_gridmap_line(gridmap_line_t *gline);
static int match_dn(const char * dn, const char * match, unsigned short matching_type);

static int match_dn(const char * match, const char * to_match, unsigned short matching_type) {
	if ( (matching_type&MATCH_WILD_CHARS) == MATCH_WILD_CHARS ){
		return fnmatch(match, to_match, 0);
	} else {
		return strcmp(match, to_match);
	}
}

gridmap_line_t * find_mapping(const gridmap_list_t * list, const char * to_match, unsigned short matching_type) {
	const gridmap_list_t * current = list;
	while (current != NULL){
		if (match_dn(current->line->match,to_match,matching_type) == 0)
			return current->line;
		current = current->next;
		if (current == list)
			break;
	}
	return NULL;
}

void gridmap_list_free(gridmap_list_t * list){
	gridmap_list_t * current = list;
	gridmap_list_t * working = NULL;
	while (current != NULL){
		working = current;
		current = current->next;

		gridmap_free_gridmap_line(working->line);
		free(working);
	}
}

int parse_gridmapfile(const char * image_map_file, gridmap_list_t ** list){
	gridmap_list_t *  result     = NULL;
	gridmap_list_t *  resultlast = NULL;
	gridmap_list_t *  resultnew  = NULL;
	gridmap_line_t *  gline      = NULL;
	char *            contents   = NULL;
	char *            linestart  = NULL;
	char *            lineend    = NULL;
	int               eof        = 0;
	int               read_rc;

	read_rc = cgul_read_config(image_map_file, &contents, getuid(), getgid());
	if (read_rc != 0){
    /*EEF_log(LOG_ERR,"Unable to safely open the image map file '%s', reason: %d (for an explenation of the reason look in cgul/fileutil/fileutil.h)",image_map_file,read);*/
		goto failure;
	}

	linestart = contents;
	while (eof == 0){
		lineend = linestart;
		while (*lineend != '\n'){
			if (*lineend == '\0'){
				eof = 1;
				break;
			}
			lineend++;
		}
		*lineend = '\0';

		read_rc = gridmap_parse_line(linestart, &gline);
		linestart = lineend+1;

		/* error which is unrecoverable only will result in a non 0 value */
		if (read_rc != 0) {
			goto failure;
		}

		/* line consisting of comment or whitespaces only */
		if (gline == NULL)
			continue;

		resultnew       = calloc(1,sizeof(gridmap_list_t));
		resultnew->line = gline;

		if (result == NULL){
			resultlast = result = resultnew;
		} else {
			resultlast->next = resultnew;
			resultlast = resultnew;
		}
		/* be sure to point the last result to the actual last result in the list */
		while (resultlast->next != NULL)
			resultlast = resultlast->next;

		gline     = NULL;
		resultnew = NULL;
	}

	*list = result;
  free(contents);

	return 0;
failure:
	gridmap_free_gridmap_line(gline);
	gline = NULL;

	free(contents);
	contents = NULL;

	gridmap_list_free(result);
	result = NULL;
	resultlast = NULL;

	return read_rc;
}

static int gridmap_parse_line(char * line, gridmap_line_t ** gline) {
	charlist_t *        mappings     = NULL;
	charlist_t *        mappingsnew  = NULL;
	char *              dn_end;
	char *              parsed_dn = NULL;
	gridmap_line_t *    gline_tmp = NULL;


	/* Check arguments */
	if ((line == NULL) || (gline == NULL))
		goto error;

	/* Skip over leading whitespace */
	line += strspn(line, WHITESPACE_CHARS);

	/* Check for comment at start of line and ignore line if present */
	if (strchr(COMMENT_CHARS, *line) != NULL) {
		/* Ignore line, return NULL gline */
		*gline = NULL;
		return 0;
	}

	/* Check for empty line */
	if (*line == NUL) {
		/* Empty line, return NULL gline. */
		*gline = NULL;
		return 0;
	}

	/* Is DN quoted? */
	if (strchr(QUOTING_CHARS, *line) != NULL) {
		/*
		 * Yes, skip over opening quote and look for unescaped
		 * closing double quote
		 */
		line++;
		dn_end = line;
		while (strchr(QUOTING_CHARS, *dn_end) == NULL){
			if (strchr(ESCAPING_CHARS, *dn_end) != NULL)
				dn_end++;
			if (*dn_end == 0)
				goto error;
			dn_end++;
		}
	} else {
		/* No, just find next whitespace */
		dn_end = line + strcspn(line, WHITESPACE_CHARS);

		if (*dn_end == NUL)
			goto error; /* Nothing after DN */
	}

	/* NUL terminate DN and parse */
	*dn_end = NUL;

	if (gridmap_parse_globusid(line, &parsed_dn) != 0)
		return -1;

	/* Skip over closing delim and any whitespace after DN */
	line = dn_end + 1;
	line += strspn(line, WHITESPACE_CHARS);

	/* Parse list of unix user ID seperated by USERID_SEP_CHARS */
	while ((*line != NUL) && (strchr(WHITESPACE_CHARS, *line) == NULL)) {
		int    mapping_len;
		char * mapping;

		/* Find end of this userid */
		mapping_len = strcspn(line, USERID_TERMINATOR_CHARS);

		mapping = strndup(line,mapping_len);
		if (mapping == NULL)
			goto error;

		mappingsnew = charlist_add_char_ptr_own(mappings, mapping);
		/* test for the memory beeing available */
		if (mappingsnew == NULL){
			/* free the local char ptr */
			free(mapping);

			goto error;
		}

		mappings = mappingsnew;

		line += mapping_len;

		/* If we're on a seperator character, skip over it */
		if (strchr(USERID_SEP_CHARS, *line) != NULL)
			line++;
	}

	/*
	 * There might be more stuff on the line, but we're ignoring
	 * it for now.
	 */

	/* Ok, build our gridmap_line_t structure */
	gline_tmp = malloc(sizeof(*gline_tmp));

	if (gline_tmp == NULL)
		goto error;

	gline_tmp->match = parsed_dn;
	gline_tmp->mappings = mappings;

	*gline = gline_tmp;

	return 0;

error:
	free(parsed_dn);
	parsed_dn = NULL;

	charlist_free(mappings);

	free(gline_tmp);

	return -1;

} /* gridmap_parse_line() */

/******************************************************************************
Function:   gridmap_parse_globusid
Description:
    Given a pointer to a string containing the globusid from the
    gridmap file, return a pointer to a string containing the
    parsed from of the id.

    Specifically handle backslashed characters - e.g. '\\',
    '\x4a' or '\37'.

Parameters:
    unparsed, pointer to unparsed string

    pparsed, pointer to pointer that should be set to point at
    allocated parsed string

Returns:
    0 on success
    non-zero on error.

******************************************************************************/
static int gridmap_parse_globusid(const char * unparsed, char ** pparsed) {
	int    escaped = 0;      /* Is the current character escaped? (Previous char was backslash) */
	char * buffer = NULL;    /* Buffer we are putting resulting name into */
	int    buffer_len = 0;   /* Buffer's length in bytes */
	int    buffer_index = 0; /* And our current position in buffer */
	char   unparsed_char;    /* Character we're currently looking at */


	/* Check input parameters for legality */
	if ((unparsed == NULL) || (pparsed == NULL))
		return -1;

	buffer_len = strlen(unparsed);
	buffer = malloc(buffer_len);
	if (buffer == NULL)
		return -1;

	/* Walk through the name, parsing as we go */
	while ((unparsed_char = *(unparsed++)) != NUL){
		/* Unescaped backslash */
		if (strchr(ESCAPING_CHARS, unparsed_char) && !escaped) {
			escaped = 1;
			continue;
		}

		/* Escaped hex character - e.g. '\xfe' */
		if ((unparsed_char == 'x') && escaped) {
			if (isxdigit(*unparsed) && isxdigit(*(unparsed + 1))) {
				/* Set unparsed_char to value represented by hex value */
				unparsed_char = (xdigit_to_value(*unparsed) << 4) + xdigit_to_value(*(unparsed + 1));
				unparsed += 2;
			}
			/* else just fall through */
		}

		/*
		 * Ok, we now have the character in unparsed_char to be appended
		 * to our output string.
		 *
		 * First, make sure we have enough room in our output buffer.
		 */
		if ((buffer_index + 1 /* for NUL */) >= buffer_len) {
			/* Grow buffer */
			char *tmp_buffer;
			buffer_len *= 2;
			tmp_buffer = realloc(buffer, buffer_len);
			if (tmp_buffer == NULL) {
				free(buffer);
				return -1;
			}
			buffer = tmp_buffer;
		}

		buffer[buffer_index++] = unparsed_char;
		buffer[buffer_index] = NUL;

		escaped = 0;
	}

	/* XXX What if escaped == 1 here? */
	/* Success */

	*pparsed = buffer;

	return 0;

} /* gridmap_parse_globusid() */

/******************************************************************************
Function:   xdigit_to_value
Description:
    Convert a ascii character representing a hexadecimal digit
    into a integer.

Parameters:
    xdigit, character contain the hex digit.

Returns:
    value contained in xdigit, or -1 on error.

******************************************************************************/

static int xdigit_to_value(char xdigit) {
	if ((xdigit >= '0') && (xdigit <= '9'))
		return (xdigit - '0');
	if ((xdigit >= 'a') && (xdigit <= 'f'))
		return (xdigit - 'a' + 0xa);
	if ((xdigit >= 'A') && (xdigit <= 'F'))
		return (xdigit - 'A' + 0xa);
	/* Illegal digit */
	return -1;
} /* xdigit_to_value() */

/******************************************************************************
Function:   gridmap_free_gridmap_line
Description:
    Frees all memory allocated to a gridmap_line_t structure.

Parameters:
    gline, pointer to structure to be freed.

Returns:
    Nothing.
******************************************************************************/
static void gridmap_free_gridmap_line(gridmap_line_t *gline) {
	if (gline == NULL)
		return;

	free(gline->match);
	gline->match = NULL;

	charlist_free(gline->mappings);
	gline->mappings = NULL;

	free(gline);
} /* gridmap_free_gridmap_line() */

void charlist_free(charlist_t * list){
	charlist_t * current = list;
	charlist_t * working;
	while (current != NULL){
		working = current;
		current = current->next;

		/* clean the element */
		free(working->value);
		free(working);
		working = NULL;

		if (current == list)
			/* it accidantly is circular so break out of it as I already cleaned it up */
			break;
	}
}

charlist_t * charlist_add_char_ptr_own(charlist_t * list, const char * str){
	charlist_t * working = list;
	charlist_t * last    = NULL;
	if (working == NULL){
		last = working = calloc(1,sizeof(charlist_t));
	} else {
		last = working;
		while (last->next != NULL){
			/* it is a circular list, now you found the last one */
			if (last->next == working)
				break;
			last = last->next;
		}
		last->next = calloc(1,sizeof(charlist_t));
		last = last->next;
	}

	if (last == NULL)
		return NULL;

	last->value = (char*)str;
	return working;
}

charlist_t * charlist_add_char_ptr_owned(charlist_t * list, const char * str) {
	char * mystr = strdup(str);
	if (mystr == NULL)
		return NULL;
	return charlist_add_char_ptr_own(list, mystr);
}

/*int main(int argc, char* argv[]){*/
/*char* gridmapfile = "/home/aczid/workspace/grid-mw-security/ees/examples/example_gridmapfile.txt";*/

/*}*/
