#include "log_to_file.h"

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <syslog.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <stdio.h>
#include <time.h>


/* example
   int main(void){
     cgul_openlog("/tmp/foo");
     cgul_set_ident("test");
     cgul_set_log_options(CGUL_SYSLOG | CGUL_LOGFILE | CGUL_STDERR);
     cgul_log(LOG_ERR, "Foo %s\n", "bar");
     cgul_closelog();
     return 0;
   }
*/

static FILE*        log_file_fp=NULL;
static char*        ident_str=NULL;
/* Number of bytes times 8*log(2)/log(10) rounded up, plus 1 extra for minus and
 * 1 extra for \0 */
static size_t const pid_str_len=sizeof(pid_t)*3+2;
static char	    pid_str[sizeof(pid_t)*3+2];

static CGUL_LOGMASK log_options;

/**
 * Internal function - Opens log file
 */
int cgul_openlog(const char* file){
  pid_t _pid = getpid();
  openlog(file, LOG_CONS | LOG_PID | LOG_NDELAY, LOG_DAEMON);
  snprintf(pid_str, pid_str_len, "%d",_pid);
/*  asprintf(&pid_str, "%llu", (long long unsigned int) _pid);*/
  if((log_file_fp = fopen(file, "a")))
    return 0;
  else
    return -1;
}

void cgul_set_ident(char* ident){
  ident_str = ident;  
}

/**
 * Internal function - Flush log file
 */
int cgul_flushlog(){
    if (log_file_fp==NULL)
	return 0;
    else
	return fflush(log_file_fp);
}

/**
 * Internal function - Closes log file
 */
int cgul_closelog(void){
/*  free(pid_str);*/
  if (log_file_fp)
    return fclose(log_file_fp);
  else
    return 0;
}

/**
 * Sets the current logging options as a bitmask
 */
void cgul_set_log_options(CGUL_LOGMASK mask){
/*  log_options |= mask;*/
  log_options = mask;
}

/**
 * Returns the current logging options as a bitmask
 */
CGUL_LOGMASK cgul_get_log_options(){
  return log_options;
}

/**
 * Main log function. Follows syslog() prototype. Depending on which log flags are set through cgul_set_log_options will log to either stderr, syslog and/or a logfile
 */
void cgul_log(int priority, const char* format, ...){
	va_list     args;
  char *      _format = append_newline(format);

  /* error logging */
  if(log_options & CGUL_STDERR){
    va_start(args, format);
    if((priority == LOG_ERR) || (priority == LOG_WARNING)){
      vfprintf(stderr, _format, args);
    }
    va_end(args);
  }

  /* debug logging */
  if(log_options & CGUL_STDDEBUG){
    va_start(args, format);
    if(priority == LOG_DEBUG){
      vfprintf(stderr, _format, args);
    }
    va_end(args);
  } else {
    /* if running in non-debug mode */
    if(priority > LOG_INFO){
      /* and priority is debug or info */
      goto end;
    }
  
  }

  /* Default output */
  if(log_options & CGUL_SYSLOG){
    va_start(args, format);
    vsyslog(priority, format, args);
    va_end(args);
  }

  if(log_options & CGUL_LOGFILE){
    va_start(args, format);
    log_to_file(priority, format, args);
    va_end(args);
  }

end:
  /* if the format string was changed by appending, it needs to be free'd */
  if(format != _format){
    free(_format);
  }

  /* if the format string was changed by stripping, it needs to be free'd */
  return;
}

/**
 * Log to file function. Follows vsyslog() prototype
 */
void log_to_file(int priority, const char* format, va_list ap){
  time_t      _time     = 0;
  struct tm * _time_s   = NULL;
  char      * _output   = NULL;
  char        _strf_fmt[MAX_TIME_STRING_SIZE];
  char      * _format   = append_newline(format);
  char        _hostname[1024];
  int         _max_format_string_length;

  if (log_file_fp==NULL) return;

  _time = time(NULL);
  _time_s = localtime(&_time);
  strftime(_strf_fmt, MAX_TIME_STRING_SIZE-1, "%b %d %H:%M:%S", _time_s);
  gethostname(_hostname, 1023);

  /* concat timestamp and format string */
  if((_output = calloc((strlen(_strf_fmt)+1+strlen(_hostname)+1+strlen(ident_str)+1+strlen(pid_str)+3+strlen(_format)+1),sizeof(char)))){
    strncpy(_output, _strf_fmt, strlen(_strf_fmt));
    strncat(_output, " ", 1);
    strncat(_output, _hostname, strlen(_hostname));
    strncat(_output, " ", 1);
    strncat(_output, ident_str, strlen(ident_str));
    strncat(_output, "[", 1);
    strncat(_output, pid_str, strlen(pid_str));
    strncat(_output, "]: ", 3);
    _max_format_string_length = (MAX_LOG_MESSAGE_SIZE - strlen(_strf_fmt) - 1 - strlen(_hostname) - 1 - strlen(ident_str) - 1 - strlen(pid_str) - 3);
    strncat(_output, _format, _max_format_string_length);
    vfprintf(log_file_fp, _output, ap);
    cgul_flushlog();
    free(_output);
  }
  if(format != _format){
    free(_format);
  }
}

/**
 * Returns a new string which is a copy of the supplied string with newline appended
 */
char* append_newline(const char* string){
  char*  mutable_string = NULL;
  size_t string_length  = strlen(string);

  if(string[string_length-1] != '\n'){
    if((mutable_string = calloc((string_length+2),sizeof(char)))){
      strncpy(mutable_string, string, string_length);
      strncat(mutable_string, "\n", 1);
    }
  } else {
    mutable_string = strdup(string);
  }
  return mutable_string; 
}

/**
 * Returns a new string which is a copy of the supplied string with the newline stripped off
 */
char* strip_newline(const char* string){
  char* mutable_string = NULL;
  size_t string_length = strlen(string);

  if((sscanf(string, "\n")) > 0){
  /*if(string[strlen(string)-1] == '\n'){*/
    if((mutable_string = calloc((string_length),sizeof(char)))){
      strncpy(mutable_string, string, string_length);
    }
  } else {
    mutable_string = strdup(string);
  }
  return mutable_string; 
}

