#include "eef_library.h"
#include "eef_xacml_parser.h"
#include "pdl.h"
#include "plugin_manager.h"

#include <unistd.h>
#include <syslog.h>

static const char* _modules_path=NULL;  /*! Holds the global path to plug-in
                                          modules */
static const char* _log_file=NULL;
static int _EES_port=-1;    /*! EES listening port */

void (*EEF_log)(int, const char*, ...);

void fork_and_exit_parent_with_logging(void){
  switch(fork()){
    case 0:
      break;
    case -1:
      EEF_log(LOG_ERR, "Error forking: %s\n", strerror(errno));
      exit(1);
    default:
      exit(0);
  }
}

unsigned int EEF_getMajorVersion(void){
  int major = 0, minor = 0, patch = 0;

  if (sscanf(VERSION, "%d.%d.%d", &major, &minor, &patch) != 3) {
    EEF_log(LOG_ERR, "%s error: could not parse compile-time version information for EEF.\n", __func__);
    return 0;
  } else
    return major;
}

unsigned int EEF_getMinorVersion(void){
  int major = 0, minor = 0, patch = 0;
  if (sscanf(VERSION, "%d.%d.%d", &major, &minor, &patch) != 3) {
    EEF_log(LOG_ERR, "%s error: could not parse compile-time version information for EEF.\n", __func__);
    return 0;
  } else
    return minor;
}

unsigned int EEF_getPatchVersion(void){
  int major = 0, minor = 0, patch = 0;
  if (sscanf(VERSION, "%d.%d.%d", &major, &minor, &patch) != 3) {
    EEF_log(LOG_ERR, "%s error: could not parse compile-time version information for EEF.\n", __func__);
    return 0;
  } else
    return patch;
}

/**
 * Tries to daemonize the process
 */
EES_RC EEF_daemonize(void){
#ifdef BSD
  pid_t pid;
#endif

  /* If launched by init there's no need to detach
   * Note: this test is unreliable due to an unavoidable race
   * condition if the process is orphaned.
   */
  if(getppid() == 1){
    goto out;
  }

#ifdef SIGTTOU
  signal(SIGTTOU, SIG_IGN);
#endif

#ifdef SIGTTIN
  signal(SIGTTIN, SIG_IGN);
#endif

#ifdef SIGTSTP
  signal(SIGTSTP, SIG_IGN);
#endif

  /* allow parent shell to continue - ensure process is not processgroup leader*/
  /*fork_and_exit_parent_with_logging();*/

#ifdef BSD
  pid = getpid();
  if(setpgid(0, pid) != -1){
    if((fd = open("/dev/tty", O_RDWR)) >=0){
      ioctl(fd, TIOCNOTTY, 0);
      close(fd);
    }
  }
#else
  setpgrp();
#endif

out:
  fork_and_exit_parent_with_logging();

  /* change dir to '/' */
  if(chdir("/") != 0){
    return EES_FAILURE; 
  }

  if (freopen("/dev/null","a",stdout)==NULL ||
      freopen("/dev/null","a",stderr)==NULL ||
      freopen("/dev/null","r",stdin)==NULL) {
      EEF_log(LOG_ERR,"Cannot reattach stdin/stderr/stdout: %s\n",strerror(errno));
      return EES_FAILURE;
  }
  
  /* set file creation mask to '750' */
  umask(027);

  return EES_SUCCESS;

}

/**
 * Try to initialize the AOS, enable logging, initialize the plug-in manager and try to parse the config file and ascertain the module path
 */
EES_RC EEF_init(char *config_file, void (*log_func)(int, const char*, ...)){
  if(log_func != NULL){
    EEF_log = log_func;
  } else {
    EEF_log = syslog;
  }

  /* FIXME this leaks memory... somehow */
  /*EEF_log(LOG_NOTICE, "EEF initializing version %i.%i.%i...",
      EEF_getMajorVersion(), EEF_getMinorVersion(), EEF_getPatchVersion());*/

  if (AOS_Init() == EES_SUCCESS){
    if(start_plugin_manager() == EES_SUCCESS){
      return start_pdl_parser(config_file, 0, NULL);
    }
  }
  return EES_FAILURE;
}

/**
 * Terminates the parser, stops the plug-in manager, terminates the AOS and closes the log.
 */
EES_RC EEF_term(){
  EEF_log(LOG_INFO, "Cleaning up...");
  /* Terminate the configuration parser */
  if((pdl_term()                    == EES_SUCCESS) &&
      /* Stop all the plugins in the plugin manager */
      (stop_plugin_manager()        == EES_SUCCESS) &&
      /* Terminate the AOS */
      (AOS_Term()                   == EES_SUCCESS)){
    /* Return with successful return code */
    return EES_SUCCESS;
  }
  /* The components should log their own errors */
  EEF_log(LOG_ERR, "EEF: Errors detected\n");
  EEF_log(LOG_NOTICE, "EEF terminating...");
  /* Return with unsuccessful return code */
  return EES_FAILURE;
}

/**
 * instructs the AOS to go into threaded mode, trying to store new data in thread-local storage
 */
EES_RC EEF_startThreading(void){
  aos_start_threading();
  plugin_manager_start_threading();
  return EES_SUCCESS;
}

/**
 * Sets the global modules path to load plug-ins from
 */
void EEF_set_modules_path(const char* path){
  _modules_path = path;
}

/**
 * Returns the global modules path to load plug-ins from
 * This can be either the _modules_path as set by the set_modules_path() function or the MODULE_DIR path macro generated by the configure script.
 */
const char* EEF_get_modules_path(void){
  if(_modules_path != NULL){
    if(strlen(_modules_path)!=0){
      return _modules_path;
    }
  }
  return MODULE_DIR;
}

void EEF_set_log_file(const char* path){
  _log_file = path;
}

/**
 * Returns the current log file
 */
const char* EEF_get_log_file(void){
  return _log_file;
  return NULL;
}

char* EEF_getRunningPluginName(void){
  return get_running_plugin_name();
}

/**
 * Sets the EES listening port
 */
void EEF_set_port(int port){
  if (port>=0)    {
    EEF_log(LOG_NOTICE, "EEF listening on port %i...",port);
    _EES_port = port;
  }
}

/**
 * Returns the EES listening port
 */
int EEF_get_port(void){
  return _EES_port;
}

