#include "pdl.h"
#include <stdlib.h>
#include "plugin_manager.h"
#include "eef_library.h"

/* Lex/Yacc stuff */
extern FILE *               yyin;
extern unsigned int         lineno;
extern int                  yyparse(void);
extern int                  yylex(void);
#if HAVE_YYLEX_DESTROY
extern int                  yylex_destroy(void);
#endif


pdl_data_t                  config;

/* This function does everything parser related. After running yacc through
 * wrap_yacc(), the created list is reduced to those in the array of named
 * policies, of number_of_policies is greater than 0. The policies are checked
 * for unreachable rules. After the parser has created a list of policies
 * containing lists of rules, these rules are transformed to a tree by the
 * link_dead_end_rules_in_policies() function. Finally, when a list of policies
 * containing rules trees was created, the plug-ins are linked to each rule in
 * the node. */
EES_RC start_pdl_parser(char *config_file, int number_of_policies, char* names_of_policies[]){
  config.pdl_path=NULL;
  config.pdl_port=-1;
  config.pdl_log_file=NULL;
  config.recursion_was_created           = 0;
  config.unknown_variable_was_referenced = 0;
  config.starting_state_was_reused       = 0;
  config.parse_errors_detected           = 0;


  EEF_log(LOG_NOTICE, "Loading config file: %s\n", config_file);
  if(wrap_yacc(config_file) == EES_SUCCESS){
    /* after the config file was successfully parsed, the policies and path to
     * plugins are accessible. We can try to set the global (EEF-wide) modules
     * path if one was provided */
    EEF_log(LOG_NOTICE, "Loaded config file: %s\n", config_file);

    /* set the EEF modules path to the path found by the evaluation manager */
    EEF_set_modules_path(get_pdl_path());

    /* set the EES port found by the evaluation manager */
    EEF_set_port(get_pdl_port());

    /* set the EEF log file to the path found by the evaluation manager */
    EEF_set_log_file(get_pdl_log_file());

    /* strip policies that are not explicitly passed in the names_of_policies array */
    if(number_of_policies){
      config.policies_list = reduce_policies(config.policies_list, number_of_policies, names_of_policies);
    }

    if(remove_unreachable_rules_in_policies(config.policies_list) == EES_SUCCESS){
      if(config.policies_list){
        link_dead_end_rules_in_policies(config.policies_list);
        link_rules_to_plugins(config.policies_list);
      }
      EEF_log(LOG_INFO, "Your configuration file seems sane.\n");
      print_policies(config.policies_list);

      /* this is a callout to the plugin manager, which dlopen()'s the plugin, dlsym()'s the plugin's functions and calls the plugin's initialize function */
      if(initialize_plugins() == EES_SUCCESS){
        return EES_SUCCESS;
      } else {
        EEF_log(LOG_ERR, "Failed to initialize plug-ins from policy config file %s\n", config_file);
      } 
    } else {
      EEF_log(LOG_ERR, "The loaded configuration file defines unreachable rules. Please check your configuration file.");
    }
  } else {
    if(config.recursion_was_created){
      EEF_log(LOG_ERR, "The loaded configuration file defines recursive rules. Please check your configuration file.");
    }
    if(config.unknown_variable_was_referenced){
      EEF_log(LOG_ERR, "The loaded configuration file references undefined variables. Please check your configuration file.");
    }
    if(config.starting_state_was_reused){
      EEF_log(LOG_ERR, "The loaded configuration file reuses a starting state. Please check your configuration file.");
    }
    if(config.parse_errors_detected){
      EEF_log(LOG_ERR, "The configuration was not parsed correctly. Please check your configuration file.");
    }
  }
  return EES_FAILURE;
}


/*! Initializes the parsing of the configuration file. When parsing has completed successfully, the policy_list is available. This can be transformed to a tree using the link_dead_end_rules_in_policies() function. */
EES_RC wrap_yacc(const char* config_file){
  config.config_file_s = config_file;
  if((config.config_file_fp = fopen(config.config_file_s,"r")) != NULL){
    yyin = config.config_file_fp;
    yyparse();
    if(!config.recursion_was_created && !config.unknown_variable_was_referenced && !config.starting_state_was_reused && !config.parse_errors_detected){
      /* callout to the plugin manager */
      if(add_plugin_structs() == EES_SUCCESS){
        return EES_SUCCESS;
      } else {
        EEF_log(LOG_ERR, "Failed to load plug-ins from policy config file %s\n", config.config_file_s);
      }
    }
  } else {
    EEF_log(LOG_ERR, "Failed to open policy config file %s", config.config_file_s);
  }
  return EES_FAILURE;
}

policy_t* get_policies(){
  return config.policies_list;
}

int check_path(char* path){
  size_t path_size = 0;
  size_t string_size;

  if(!path){
    return 0;
  }

  string_size = strlen(path);

  /* TODO add a stat here */

  return path_size;
}

/*! sets the path to modules directory */
void set_pdl_path(record_t* path){
  size_t path_size = 0;
  size_t string_size;

  if(!path || !path->string){
    return;
  }

  string_size = strlen(path->string)+2; /* size of string + extra slash + null byte */
  if(string_size < FILENAME_MAX){
    path_size = string_size;
  } else {
    path_size = FILENAME_MAX;
  }

  if((config.pdl_path = calloc(1, path_size))){
    snprintf(config.pdl_path, path_size, "%s/", path->string);
  }

  EEF_log(LOG_DEBUG, "Found a new modules path: %s\n", config.pdl_path);

  free(path->string);
  free(path);
}

/*! sets the path to the logfile */
void set_pdl_log_file(record_t* path){
  size_t path_size = 0;
  size_t string_size;
  if(!path){
    return;
  }

  string_size  = strlen(path->string)+1; /* size of string + null byte */

  if(string_size < FILENAME_MAX){
    path_size = string_size;
  } else {
    path_size = FILENAME_MAX;
  }

  if((config.pdl_log_file = calloc(1, path_size))){
    snprintf(config.pdl_log_file, path_size, "%s", path->string);
  }

  EEF_log(LOG_DEBUG, "Found a new logfile path: %s\n", config.pdl_log_file);

  free(path->string);
  free(path);
}

/*! sets the EES listening port */
void set_pdl_port(record_t* port){
  EEF_log(LOG_DEBUG, "Found port: %s\n", port->string);

  /*struct stat sb;*/
  /*if stat(MODUL*/
  if (sscanf(port->string,"%d",&(config.pdl_port))!=1)    {
      EEF_log(LOG_ERR, "Cannot parse port number: %s\n", port->string);
      config.pdl_port=-1;
  } else {
      EEF_log(LOG_DEBUG, "Using listening port: %d\n", config.pdl_port);

      free(port->string);
      port->string = NULL;
      free(port);
      port = NULL;
  }
}

/*! Handles a variable representing a plug-in; creates a structure containing its values and adds it to the list of var_t structs variables_list */
void add_variable(record_t* name, record_t* value){
  EEF_log(LOG_DEBUG, "Added variable name: %s\n", name->string);
  EEF_log(LOG_DEBUG, "Added variable value: %s\n", value->string);

  /* allocate struct and populate fields */
  if((config.current_variable = calloc(1,sizeof(var_t)))){
    config.current_variable->name = strdup(name->string);
    config.current_variable->value = strdup(value->string);
    config.current_variable->lineno = name->lineno;
    config.current_variable->next = NULL;
  } else {
    EEF_log(LOG_ERR, "Out of memory!");
  }

  /* append to the end of the list */
  if(config.variables_list){
    config.variables_list_last->next = config.current_variable;
  } else {
    config.variables_list = config.current_variable;
  }
  config.variables_list_last = config.current_variable;

  /* clean up */
/*  free(name->string);
  name->string = NULL;
  free(name);
  name = NULL;*/
  free(value->string);
  value->string = NULL;
  free(value);
  value = NULL;
}

/* TODO I think this function is a bit of a kludge and could probably be rewritten to be more efficient and more legible - I think it might do too much, so extracting some functions might help */
/*! Appends a rule to the global rules_list, which is to be added to a policy when add_policy is called
 Will return NULL and log an error when:
 - adding the rule to the list will create a recursion in the tree of rules
 - an unknown variable is referenced
*/
rule_t* add_rule(record_t* state, record_t* true_branch, record_t* false_branch){
  rule_t *new_rule = NULL, *new_false_branch = NULL, *new_true_branch = NULL, *recursive_rule = NULL;
  var_t  *temp_var = NULL;
  char *unknown_var_format_string = "Unknown variable %s at line %i in config file %s";

  if((new_rule = get_left_hand_rule(config.rules_list, state->string))){
    EEF_log(LOG_WARNING, "State %s at line %i is already in use at line %i.\n", state->string, state->lineno, new_rule->lineno);
    config.starting_state_was_reused = 1;
  } else {
    /*find variables for rule */
    temp_var = get_variable_by_name(state->string);
    if(temp_var == NULL){
      /* Errorous state - variable referenced in rule not previously defined */
      EEF_log(LOG_ERR, unknown_var_format_string, state->string, state->lineno, config.config_file_s);
      config.unknown_variable_was_referenced = 1;
      new_rule = clean_rules_tree(new_rule);
      goto cleanup;
    }
    if(temp_var != NULL){
      if((new_rule = calloc(1, sizeof(rule_t)))){
        /* populate fields of current state */
        new_rule->name = strdup(state->string);
        new_rule->lineno = state->lineno;

        /* populate fields of branches */
        if(false_branch){
          temp_var = get_variable_by_name(false_branch->string);
          if(temp_var == NULL){
            /* Errorous state - variable referenced in rule not previously defined */
            EEF_log(LOG_ERR, unknown_var_format_string, false_branch->string, false_branch->lineno, config.config_file_s);
            config.unknown_variable_was_referenced = 1;
            new_rule = clean_rules_tree(new_rule);
            goto cleanup;
          } else {
            if((new_false_branch = calloc(1, sizeof(rule_t)))){
              new_false_branch->name = strdup(false_branch->string);
              new_false_branch->lineno = false_branch->lineno;
              new_rule->false_branch = new_false_branch;
            }
          }
        }

        if(true_branch){
          temp_var = get_variable_by_name(true_branch->string);
          if(temp_var == NULL){
            /* Errorous state - variable referenced in rule not previously defined */
            EEF_log(LOG_ERR, unknown_var_format_string, true_branch->string, true_branch->lineno, config.config_file_s);
            config.unknown_variable_was_referenced = 1;
            new_rule = clean_rules_tree(new_rule);
            goto cleanup;
          } else {
            if((new_true_branch = calloc(1, sizeof(rule_t)))){
              new_true_branch->name = strdup(true_branch->string);
              new_true_branch->lineno = true_branch->lineno;
              new_rule->true_branch = new_true_branch;
            }
          }
        }

        /* check for recursion */
        if((recursive_rule = check_for_recursion(config.rules_list, new_rule))){
          EEF_log(LOG_WARNING, "Rule %s at line %i leads to recursion into state %s", new_rule->name, new_rule->lineno, recursive_rule->name);
          new_rule = clean_rules_tree(new_rule);
          config.recursion_was_created = 1;
          goto cleanup;
        } else {
          /* add new rule at the end of the rules list */
          if(config.rules_list){
            config.rules_list_last->next = new_rule;
          } else {
            config.rules_list = new_rule;
          }
          config.rules_list_last = new_rule;

          EEF_log(LOG_DEBUG, "Added a new rule: %s\n", new_rule->name);
        }

      } else {
        EEF_log(LOG_ERR, "Out of memory!");
      }
    }
  }

cleanup:

  /* clean up */
  if(state != NULL){
    free(state->string);
    free(state);
  }
  if(true_branch != NULL){
    free(true_branch->string);
    free(true_branch);
  }
  if(false_branch != NULL){
    free(false_branch->string);
    free(false_branch);
  }

  return new_rule;
}

/* Tries to find a recursion in the list of rules by iterating through all the tree branches. When the same rule is encountered in the current list a recursion has been found */
rule_t* check_for_recursion(rule_t* rule_l, rule_t* rule_r){
  rule_t* temp_rule = rule_l;

  /* right hand rule leads to its own recursive state */
  /* still need to find a way to reuse the code from the while loop to do this check */
  if(rule_r){
    if(rule_r->true_branch){
      if(strcmp(rule_r->name, rule_r->true_branch->name) == 0){
        return rule_r->true_branch;
      }
    }

    if(rule_r->false_branch){
      if(strcmp(rule_r->name, rule_r->false_branch->name) == 0){
        return rule_r->false_branch;
      }
    }
  }

  /* iterate list of rules */
  while(temp_rule){
    if((temp_rule != NULL) && (rule_r != NULL)){
      /* left hand and right hand state are equal */
      if(strcmp(temp_rule->name, rule_r->name) == 0){
        return rule_r;
      }

      /* start state and true branch of right hand rule are equal */
      if(rule_r->true_branch){
        if(strcmp(temp_rule->name, rule_r->true_branch->name) == 0){
          return rule_r->true_branch;
        }
      }

      /* start state and false branch of right hand side are equal */
      if(rule_r->false_branch){
        if(strcmp(temp_rule->name, rule_r->false_branch->name) == 0){
          return rule_r->false_branch;
        }
      }
    }

    /* move to next rule */
    temp_rule = temp_rule->next;
  }
  return NULL;
}

/* This function iterates the passed list of policies and removes unreachable rules from each policy */
EES_RC remove_unreachable_rules_in_policies(policy_t* policies){
  policy_t *temp_policy = policies;
  EES_RC retval = EES_SUCCESS;
  while(temp_policy){
    temp_policy->rules = remove_unreachable_rules(temp_policy->rules, &retval);
    temp_policy = temp_policy->next;
  }
  return retval;
}

/* This function iterates the passed list of rules and removes those rules that, while they exist on the left-hand side, are never referenced on the right-hand side of a rule in the list */
rule_t* remove_unreachable_rules(rule_t* rules, EES_RC *retval){
  rule_t *temp_rule = NULL, *next_rule = NULL, *previous_rule = rules;
  if(rules){
    temp_rule = rules;
    while(temp_rule){
      next_rule = temp_rule->next;

      if(!(get_left_hand_rule(rules, temp_rule->name)) && !(get_right_hand_rule(rules, temp_rule->name))){
        EEF_log(LOG_WARNING, "Removing unreachable rule %s at line %i\n", temp_rule->name, temp_rule->lineno);
        clean_rules_tree(temp_rule);
        previous_rule->next = next_rule;
        temp_rule = previous_rule;
        *retval = EES_FAILURE;
      }
      previous_rule = temp_rule;
      temp_rule = next_rule;
    }
  }
  return rules;
}

void link_dead_end_rules_in_policies(policy_t* policy){
  policy_t* temp_policy = policy;
  while(temp_policy){
    link_dead_end_rules(temp_policy, temp_policy->rules);
    temp_policy = temp_policy->next;
  }
}

void link_dead_end_rules(policy_t* policy, rule_t* rule){
  rule_t *temp_rule = rule, *true_rule, *false_rule;
  while(temp_rule){
    if(temp_rule->true_branch){
      if((true_rule = get_left_hand_rule(policy->rules, temp_rule->true_branch->name))){
        EEF_log(LOG_DEBUG, "Overwriting rule %s with %s\n", temp_rule->true_branch->name, true_rule->name);
        clean_rules_tree(temp_rule->true_branch);
        temp_rule->true_branch = true_rule;
      }
    }
    if(temp_rule->false_branch){
      if((false_rule = get_left_hand_rule(policy->rules, temp_rule->false_branch->name))){
        EEF_log(LOG_DEBUG, "Overwriting rule %s with %s\n", temp_rule->false_branch->name, false_rule->name);
        clean_rules_tree(temp_rule->false_branch);
        temp_rule->false_branch = false_rule;
      }
    }
    temp_rule = temp_rule->next; 
  }
  policy->rules_list_transformed_to_tree = 1;
}

/*! Tries to find specified state in the rules_list */
rule_t* get_left_hand_rule(rule_t* temp_rule, const char* name){
  if(!temp_rule || !name){
    return NULL;
  }

  /* iterate while the rule isn't associated with the given state */
  while(temp_rule){
    if(!strcmp(name, temp_rule->name)){
      return temp_rule;
    }
    temp_rule = temp_rule->next;
  }
  return NULL;
}

/*! Tries to find specified state in the rules_list */
rule_t* get_right_hand_rule(rule_t* temp_rule, const char* name){
  if(!temp_rule || !name){
    return NULL;
  } else {
  /* iterate while the rule isn't associated with the given state */
    if(temp_rule->true_branch){
      if(strcmp(name, temp_rule->true_branch->name) == 0){
        return temp_rule->true_branch;
      }
    }
    if(temp_rule->false_branch){
      if(strcmp(name, temp_rule->false_branch->name) == 0){
        return temp_rule->false_branch;
      }
    }
  }
  return NULL;
}

/*! Adds a new policy_t structure to the policies_list which will hold the current rules_list. rules_list is nulled after creating a policy so a new list can be created for the next policy */
void add_policy(record_t* policy, rule_t* top_rule){
  policy_t* new_policy = NULL;
  /*eef_log(LOG_DEBUG, "Found a new policy: %s\n", policy->string);*/

  if((new_policy = calloc(1, sizeof(policy_t)))){
    new_policy->name = strdup(policy->string);
    new_policy->lineno = policy->lineno;
    new_policy->rules = top_rule;
    new_policy->rules_list_transformed_to_tree = 0;
    new_policy->next = NULL;
  }

  /* append to the end of the list */
  if(config.policies_list){
    config.policies_list_last->next = new_policy;
  } else {
    config.policies_list = new_policy;
  }
  config.policies_list_last = new_policy;

  /* start a new rules list */
  config.rules_list = NULL;

  free(policy->string);
  policy->string = NULL;
  free(policy);
  policy = NULL;
}

/* prints the list of policies */
void print_policies(policy_t* policies){
  policy_t* temp_policy = policies;
  while(temp_policy){
    EEF_log(LOG_DEBUG, "Policy: %s\n", temp_policy->name);
    print_rules(temp_policy, temp_policy->rules);
    temp_policy = temp_policy->next;
  }
  return;
}


/* prints the list of rules */
void print_rules(policy_t* policy, rule_t* rule){
  if(rule){
    EEF_log(LOG_DEBUG, "  | Rule %s\n", rule->name);
    EEF_log(LOG_DEBUG, "  -------------------------\n");
    if(rule->true_branch){
      print_rules(policy, get_left_hand_rule(policy->rules, rule->true_branch->name));
    }
    if(rule->false_branch){
      print_rules(policy, get_left_hand_rule(policy->rules, rule->false_branch->name));
    }
  } 
  return;
}

/* tries to link all policy rules to a plugin in the plugin manager */
EES_RC link_rules_to_plugins(policy_t* policies){
  policy_t *temp_policy = policies;

  while(temp_policy){
    link_rule_to_plugin(temp_policy, temp_policy->rules);
    temp_policy = temp_policy->next;
  }
  return EES_SUCCESS;
}

void link_rule_to_plugin(policy_t *policy, rule_t *rule){
  var_t    *temp_var    = NULL;
  char**    argv;
  int       argc;

  eef_plugindl_t *temp_plugin = NULL;
  if(rule){
    if(!rule->plugin){
      temp_var = get_variable_by_name(rule->name);
      if(temp_var){
        argv = _var_to_argv(temp_var->value, &argc);

        temp_plugin = get_plugin(temp_var->name);
        if(temp_plugin){
          rule->plugin = temp_plugin;
        } else {
          EEF_log(LOG_WARNING, "Unknown plugin in variable %s\n", temp_var->value);
        }
        free_args(argc, argv);
      }
    } else {
      EEF_log(LOG_DEBUG, "Plugin already linked?\n");
    }
    link_rule_to_plugin(policy, rule->true_branch);
    link_rule_to_plugin(policy, rule->false_branch);
  }
  return;
}

/* iterates the policies and removes those that are not explicitly named in the passed array of strings */
policy_t* reduce_policies(policy_t* policies, int number_of_policies, char* names_of_policies[]){
  int i, policy_should_be_removed;
  policy_t *temp_policy = policies, *next_policy = NULL;

  while(temp_policy){
    policy_should_be_removed = 1;
    EEF_log(LOG_DEBUG, "Checking policy: %s\n", (temp_policy)->name);
    for(i = 0; i < number_of_policies; i++){
      if(strcmp(temp_policy->name, names_of_policies[i]) == 0){
        /* if the policy is in the named list, we can continue to the next plugin */
        EEF_log(LOG_DEBUG, "Allowed policy: %s\n", names_of_policies[i]);
        policy_should_be_removed = 0;
        break;
      }
    }

    next_policy = temp_policy->next;

    if(policy_should_be_removed){
      EEF_log(LOG_DEBUG, "Removing not-allowed policy: %s\n", temp_policy->name);
      policies = remove_policy_by_name(policies, temp_policy->name);
    }
    temp_policy = next_policy;
  }

  return policies;
}

/*! concatenates two strings */
record_t* concat_strings(record_t* r1, record_t* r2){
  record_t* new_record;
  /*eef_log(LOG_DEBUG, "Concating: %s with %s\n", r1->string, r2->string);*/
  if((new_record = malloc(sizeof(record_t)))){
    if((new_record->string =
        calloc(1, sizeof(char)*(strlen(r1->string) + strlen(r2->string)+1)))){
      strncat(new_record->string, r1->string, strlen(r1->string));
      strncat(new_record->string, r2->string, strlen(r2->string));
      free(r1->string);
      free(r1);
      free(r2->string);
      free(r2);
      return new_record;
    }
  }
  return NULL;
}

/*! concatenates two strings with a space in between */
record_t* concat_strings_with_space(record_t *r1, record_t* r2){
  record_t *r;
  /*eef_log(LOG_DEBUG, "Concating: %s with %s with spaces\n", r1->string, r2->string);*/
  if((r = malloc(sizeof(record_t)))){
      r->string = calloc(1,(sizeof(char)*(strlen(r1->string)+strlen(r2->string)+2)));
      strncat(r->string, r1->string, strlen(r1->string));
      strncat(r->string, " ", 1);
      strncat(r->string, r2->string, strlen(r2->string));
      free(r1->string);
      free(r1);
      free(r2->string);
      free(r2);
    return r;
  }
  return NULL;
}

/*! logs errors encountered during parsing */
int yyerror(char* string){
  config.parse_errors_detected = 1;
  EEF_log(LOG_ERR, "Parse error: %s at line %i in config file %s", string, lineno, config.config_file_s);
  return 0;
}

/*! get variable from list */
var_t* get_variable_by_name(char* name){
  for(config.current_variable = config.variables_list; config.current_variable != NULL; config.current_variable = config.current_variable->next){
    if(!strncmp(name, config.current_variable->name, strlen(config.current_variable->name))){
      return config.current_variable;
    }
  }
  return NULL;
}

/*! Removes a policy from the list of policies */
void remove_policy(record_t* policy){
  policy_t *temp_policy = config.policies_list, *next_policy = NULL;
  EEF_log(LOG_DEBUG, "Deleted policy: %s\n", policy->string);

  while(temp_policy){
    if(strcmp(temp_policy->name, policy->string) == 0){
      /* target policy was found */
      next_policy = temp_policy->next;

      /* Clean policy fields */
      if(temp_policy->rules_list_transformed_to_tree){
        temp_policy->rules = clean_rules_tree(temp_policy->rules);
      } else {
        temp_policy->rules = clean_rules_list(temp_policy->rules);
      }
      free(temp_policy->name);
      temp_policy->name = NULL;
      free(temp_policy);
      temp_policy = NULL;
    }

    /* Move to next policy */
    temp_policy = next_policy;
  }

  free(policy->string);
  free(policy);
  return;
}

/*! Removes a policy from the list of policies */
policy_t* remove_policy_by_name(policy_t* policies, char* policy){
  policy_t *temp_policy = NULL, *next_policy = NULL, *previous_policy = NULL;

  for(temp_policy = policies; temp_policy != NULL; ){
    next_policy = temp_policy->next;

    if(strcmp(temp_policy->name, policy) == 0){
      /* target policy was found */

      /* Clean policy fields */
      if(temp_policy->rules_list_transformed_to_tree){
        clean_rules_tree(temp_policy->rules);
      } else {
        clean_rules_list(temp_policy->rules);
      }

      EEF_log(LOG_DEBUG, "Deleted policy: %s\n", policy);
      free(temp_policy->name);
      free(temp_policy);

      /* if the head node is deleted, overwrite it with the next node */
      if(previous_policy){
        previous_policy->next = next_policy;
      } else {
        policies = next_policy;
      }
      break;
    }

    previous_policy = temp_policy;

    /* Move to next policy */
    temp_policy = next_policy;
  }

  return policies;
}

/*! converts a string to an array of strings by splitting it at each \t delimiter
  - overwrites the second argument with a pointer to the number of elements in the array */
char** _var_to_argv(char* value, int *argc){
  char   *start_of_arg = NULL,  *copy_of_value = NULL, **argv = NULL;
  char   *delimiters   = " \t";
  size_t size_of_arg   = 0,     size_of_array,         i      = 0;
  char   *str_ptr;

  copy_of_value = strdup(value);
  size_of_array = (sizeof(char*)*(strlen(copy_of_value)+1));

  if((argv = calloc(1, size_of_array)) != NULL){
    start_of_arg = strtok_r(copy_of_value, delimiters, &str_ptr);
    while(start_of_arg != NULL){
      size_of_arg = (sizeof(char)*(strlen(start_of_arg)+1));
      if((argv[i] = calloc(1, size_of_arg))){
        memcpy(argv[i], start_of_arg, size_of_arg);
        start_of_arg = strtok_r(NULL, delimiters, &str_ptr);
        i++;
      }
    }
  }
  free(copy_of_value);
  copy_of_value = NULL;
  if(i < ARG_MAX){
    *argc = i;
  } else {
    *argc = ARG_MAX;
  }
  return argv;
}

/*! returns pdl path found in parsed config file */
const char* get_pdl_path(){
  return config.pdl_path;
}

/*! returns pdl port found in parsed config file */
int get_pdl_port(){
  return config.pdl_port;
}

const char* get_pdl_log_file(){
  return config.pdl_log_file;
}


/*! Iterates list of policies and the rules they contain and tries to let the plugin manager prepare plugin structs*/
EES_RC add_plugin_structs(){
  char**    argv;
  int       argc;
  EES_RC    plugin_struct_added_ok = EES_SUCCESS;
  var_t*    temp_var = config.variables_list;

  while((temp_var != NULL) && (plugin_struct_added_ok == EES_SUCCESS)){
    argv = _var_to_argv(temp_var->value, &argc);

    /* this is a callout to the plug-in manager, which adds a struct describing a single plug-in to its list */
    plugin_struct_added_ok = add_plugin_struct(argc, argv, temp_var->name);

    /* Move to next variable */
    temp_var = temp_var->next;
  }

  return plugin_struct_added_ok;
}

/*! Iterates the list of var_t structures and tries to free them */
EES_RC clean_variables_list(){
  config.current_variable = config.variables_list;
  while(config.current_variable){
    free(config.current_variable->name);
    config.current_variable->name = NULL;
    free(config.current_variable->value);
    config.current_variable->value = NULL;

    /* Move to next variable */
    config.variables_list_last = config.current_variable;
    config.current_variable = config.current_variable->next;

    /* Clean last variable struct */
    free(config.variables_list_last);
    config.variables_list_last = NULL;
  } 
  return EES_SUCCESS;
}

/*! Iterates the tree of rule_t structures starting with the passed rule and tries to free them and their true/false branches */
rule_t* clean_rules_tree(rule_t* top_rule){
  rule_t* temp_rule = top_rule;
  if(temp_rule){
    clean_rules_tree(temp_rule->true_branch);

    clean_rules_tree(temp_rule->false_branch);

    free(temp_rule->name);
    free(temp_rule);
    temp_rule = NULL;
  } 
  return NULL;
}

/*! Iterates the list of rule_t structures starting with the passed rule and tries to free them and their true/false branches */
rule_t* clean_rules_list(rule_t* top_rule){
  rule_t *temp_rule = top_rule, *next_rule = NULL;
  while(temp_rule){
    next_rule = temp_rule->next;
    clean_rules_tree(temp_rule);
    temp_rule = next_rule;
  } 
  return NULL;
}

void free_args(int argc, char** argv){
  for(; argc > 0; argc--){
    free(argv[argc-1]);
    argv[argc-1] = NULL;
  }
  free(argv);
  argv = NULL;
}

/*! Iterates the list of policy_t structures and tries to free them and their rules */
EES_RC clean_policies_list(policy_t* policies){
  policy_t *temp_policy = policies, *last_policy = NULL;
  while(temp_policy){
    /* Clean policy fields */
    if(temp_policy->rules_list_transformed_to_tree){
      clean_rules_tree(temp_policy->rules);
    } else {
      clean_rules_list(temp_policy->rules);
    }
    free(temp_policy->name);
    temp_policy->name = NULL;

    /* Move to next policy */
    last_policy = temp_policy;
    temp_policy = temp_policy->next;
    free(last_policy);
    last_policy = NULL;
  }
  return EES_SUCCESS;
}

/*! Terminates the parser and tries to free all used memory */
EES_RC pdl_term(){
  clean_variables_list();
  clean_policies_list(config.policies_list);

  free(config.pdl_path);
  config.pdl_path = NULL;

  /*free(config.pdl_log_file);*/
  /*config.pdl_log_file = NULL;*/

  config.recursion_was_created = 0;
  config.unknown_variable_was_referenced = 0;
  config.starting_state_was_reused = 0;
  config.parse_errors_detected = 0;

  config.variables_list = NULL;
  config.variables_list_last = NULL;
  config.current_variable = NULL;

  config.rules_list = NULL;
  config.rules_list_last = NULL;

  config.policies_list = NULL;
  config.policies_list_last = NULL;

  if(config.config_file_fp != NULL){
    /* This is the preferred way of cleaning up various flex versions - See: http://flex.sourceforge.net/manual/Memory-leak-_002d-16386-bytes-allocated-by-malloc_002e.html#Memory-leak-_002d-16386-bytes-allocated-by-malloc_002e */
    #if HAVE_YYLEX_DESTROY /* macro set by configure script */
      yylex_destroy();
    #else
      /*eef_log(LOG_WARNING, "Lex function yylex_destroy() not available - possibly unable to free allocated memory for evaluation manager\n");*/
      #if HAVE_FLEX
        delete_lex_buffer(); /* this function is defined in pdl_lex.l */
        /*eef_log(LOG_INFO, "Managed to free allocated memory for evaluation manager\n"); */
      #else
        EEF_log(LOG_WARNING, "Lex functions yylex_destroy() and yy_delete_buffer() are both not available - unable to free allocated memory for evaluation manager\n");
      #endif
    #endif

    if((fclose(config.config_file_fp)==0) ){
      return EES_SUCCESS;
    }
  }
  return EES_FAILURE;
}

/*! TODO */
EES_RC allow_rules(int val){
  /*eef_log(LOG_DEBUG, "Allowed rule: %i\n", val);*/
  return EES_SUCCESS;
}

/*! TODO */
EES_RC allowed_policy_rule(const char* label){
  /*eef_log(LOG_DEBUG, "Allowed policy rule: %s\n", label);*/
  return EES_SUCCESS;  
}
