#include "plugin_manager.h"
#include "_plugin_manager.h"
#include "eef_library.h"

static pthread_key_t           _plugin_key;
static pthread_key_t           _plugin_name_key;
static pthread_once_t  	       _plugin_key_once = PTHREAD_ONCE_INIT; /* Normally not okay to do initialisation outside of a function, but this way it (hopefully) will only be set one - during compile time */
static char                    abs_plugin_path[FILENAME_MAX];
static init_fnc_ptr_t          init_fnc_ptr;
static run_fnc_ptr_t           run_fnc_ptr;
static term_fnc_ptr_t          term_fnc_ptr;

static struct eef_plugindl_s * _plugin_list;

/**
 * Starts the plugin manager (only initializes values to NULL)
 */
EES_RC start_plugin_manager(){
  _plugin_list = NULL;
  _global_running_plugin = NULL;
  pthread_once(&_plugin_key_once, plugin_manager_make_key);
  return EES_SUCCESS;
}

/**
 * Prepares the supplied plugin by name, initializes the struct with these values
 * \post the plugin node is ready to be passed to the initialize_plugin function
 */
eef_plugindl_t * create_plugin_struct(int argc, char** argv, char* name){
	eef_plugindl_t * current_plugin = NULL;
  int i = 0;

	/* current plugin struct */
	if((current_plugin = calloc(1,sizeof(eef_plugindl_t)))){

    strncpy(current_plugin->name, name, FILENAME_MAX);
    current_plugin->init_argc = argc;
    for(i = 0; i < argc; i++){
      current_plugin->init_argv[i] = (char*) strdup(argv[i]);
    }
    current_plugin->next = NULL;

    /* cleanup */
    for(i = 0; i < argc; i++){
      free(argv[i]);
    }
    free(argv);

    /* print information */
    /*print_eef_plugin(LOG_DEBUG, current_plugin);*/

    /* return plugin struct */
    return current_plugin;
  }
  return NULL;
}

/**
 * Calls create_plugin_struct() to create a plugin struct and adds it to the plugin_list
 */
EES_RC add_plugin_struct(int argc, char* argv[], char* name){
  eef_plugindl_t *last_plugin = NULL, *current_plugin = NULL;
  /* create_plugin_struct returns NULL if plugin failed to load. */
  if((current_plugin = create_plugin_struct(argc, argv, name)) != NULL){
    /* append plugin to the end of the list */
    if(_plugin_list == NULL){
      _plugin_list = current_plugin;
    } else {
      last_plugin = _plugin_list;
      /* TODO don't iterate list */
      while(last_plugin->next != NULL){
        last_plugin = last_plugin->next;
      }
      last_plugin->next = current_plugin;
    }
    /*print_eef_plugin(LOG_INFO, current_plugin);*/
    return EES_SUCCESS;
  }
  return EES_FAILURE;
}

/**
 * Wrapper for dlsym with error logging. Adapted from lcmaps_pluginmanager.c get_procsymbol
 */
eef_proc_t get_procsymbol(void* handle, char* symname){
	eef_proc_t symhandle;
  char* errstring;

  /* dlsym returns void pointer (data pointer) whereas we need to cast to a function pointer.
     This is a known problem. See: http://www.opengroup.org/onlinepubs/009695399/functions/dlsym.html#tag_03_112_08
     Although gcc will produce a warning this seems to work on most platforms
  */
	symhandle = dlsym(handle,symname);

	if((errstring = dlerror()) != NULL){
		EEF_log(LOG_ERR, "ees.get_procsymbol(): dlsym error %s\n", errstring);
		return NULL;
	}

	return symhandle;
}

/**
 * Returns a plugin from the list by name, or NULL if not found
 */
eef_plugindl_t* get_plugin(char* plugin_name){
	eef_plugindl_t * current_plugin = _plugin_list;
  /* iterate plugin_list  until plugin is found */
	while(current_plugin != NULL){
		if(strcmp(current_plugin->name, plugin_name) == 0){
			return current_plugin;
		}
		current_plugin = current_plugin->next;
	}
	return NULL;
}

/**
 * Initializes all the plugins
 * Returns EES_SUCCESS or EES_FAILURE if a plugin failed to load.*/
EES_RC initialize_plugins(){
  eef_plugindl_t* node          = _plugin_list;
  local_module_dir              = EEF_get_modules_path();
  EEF_log(LOG_DEBUG, "Using modules path: %s\n", local_module_dir);
  while(node != NULL){
    set_running_plugin_name(node->name);
    EEF_log(LOG_DEBUG, "Initializing: %s\n", node->name);
    if(initialize_plugin(node) == EES_FAILURE){
      return EES_FAILURE;
    }
    node = node->next;
  }
  return EES_SUCCESS;
}

/**
 * Initializes the supplied plugin node, which already holds the plugin name and arguments
 * \post the plugin node is initialized and its functions are linked and accessible
 */
EES_RC initialize_plugin(eef_plugindl_t* plugin){
  size_t max_plugin_name = FILENAME_MAX - strlen(local_module_dir);

  if(strlen(local_module_dir) >= FILENAME_MAX){
    return EES_FAILURE;
  }
	memset(abs_plugin_path, 0, FILENAME_MAX);

	/*TODO make safer */
	if(!(strncat(abs_plugin_path, local_module_dir, strlen(local_module_dir)) && \
    strncat(abs_plugin_path, "/", 1) && \
		strncat(abs_plugin_path, plugin->init_argv[0], max_plugin_name) 
    /*&& strcat(abs_plugin_path, ".mod")*/
    )){
    return EES_FAILURE;
	}

  EEF_log(LOG_DEBUG, "Loading plugin %s\n", abs_plugin_path);

  /* load plugin handle */
  if(!(plugin->handle = dlopen(abs_plugin_path, RTLD_LAZY))){
    EEF_log(LOG_ERR, "Failed to acquire handle on plugin: %s\n", abs_plugin_path);
    EEF_log(LOG_ERR, "%s\n", dlerror());
    return EES_FAILURE;
  }

  EEF_log(LOG_DEBUG, "Linking plugin: %s\n", abs_plugin_path);

  plugin->procs[INITPROC] = NULL;
  plugin->procs[RUNPROC] = NULL;
  plugin->procs[TERMPROC] = NULL;

  /* link function pointers */
  plugin->procs[INITPROC] = get_procsymbol(plugin->handle, "plugin_initialize");
  if(plugin->procs[INITPROC] == NULL){
    EEF_log(LOG_ERR,"ees.mod-PluginInit(): plugin %s not compliant\n", plugin->name);
    return EES_FAILURE;
  }

  plugin->procs[RUNPROC] = get_procsymbol(plugin->handle, "plugin_run");
  if(plugin->procs[RUNPROC] == NULL){
    EEF_log(LOG_ERR, "ees.runPlugin\n");
    return EES_FAILURE;
  }

  plugin->procs[TERMPROC] = get_procsymbol(plugin->handle, "plugin_terminate");
  if(plugin->procs[TERMPROC] == NULL){
    EEF_log(LOG_ERR, "ees.termPlugin\n");
    return EES_FAILURE;
  }

  set_running_plugin(plugin);
  EEF_log(LOG_DEBUG, "Initializing plugin %s\n", plugin->name);
  init_fnc_ptr = (init_fnc_ptr_t)plugin->procs[INITPROC];

  /* call function pointer */
  if(init_fnc_ptr(plugin->init_argc, plugin->init_argv) == EES_PL_SUCCESS){
    return EES_SUCCESS;
  } else {
    EEF_log(LOG_ERR, "Failed to initialize plugin %s\n", abs_plugin_path);
  }

  memset(abs_plugin_path, 0, strlen(abs_plugin_path) + 1);

  return EES_FAILURE;
}

/**
 * Runs all the plugins
 * Returns EES_SUCCESS if a plugin branch executed successfully. Returns EES_FAILURE if all plugin branches were exhausted unsuccessfully */
EES_RC run_plugins(policy_t* policy){
  rule_t* temp_rule;
  EES_RC retval = EES_FAILURE;
  /* iterate policies */
  while(policy && retval == EES_FAILURE){
    /* evaluate rules */
    temp_rule = policy->rules;
    while(temp_rule){
      EEF_log(LOG_DEBUG, "Evaluating rule: %s\n", temp_rule->name);
      if(temp_rule->plugin){
        set_running_plugin_name(temp_rule->plugin->name);
        if(run_plugin(temp_rule->plugin) == EES_PL_SUCCESS){
          if(temp_rule->true_branch){
            EEF_log(LOG_DEBUG, "Progressing to true branch %s\n", temp_rule->true_branch->name);
            temp_rule = temp_rule->true_branch;
            continue;
          } else {
            EEF_log(LOG_DEBUG, "Executed policy %s successfully\n", policy->name);
            EEF_log(LOG_DEBUG, "Ended in rule %s\n", temp_rule->name);
            /* all rules were exhausted and a successful state was reached */
            retval = EES_SUCCESS;
            break;
          }
        } else {
          if(temp_rule->false_branch){
            EEF_log(LOG_DEBUG, "Progressing to false branch %s\n", temp_rule->false_branch->name);
            temp_rule = temp_rule->false_branch;
            continue;
          } else {
            /* all rules were exhausted and no successful state was reached */
            retval = EES_FAILURE;
            break;
          }
        }
      } else {
        /* plugin not linked to current rule - This should never happen */
        EEF_log(LOG_WARNING, "Rule %s at line %i is not linked to a plugin!\n", temp_rule->name, temp_rule->lineno);
        return EES_FAILURE;
      }
    }
    policy = policy->next;
  }

  return retval;
}

/**
 * Calls the plugin's run method
 */
EES_PL_RC run_plugin(eef_plugindl_t* plugin){
  run_fnc_ptr = (run_fnc_ptr_t) plugin->procs[RUNPROC];
  EEF_log(LOG_DEBUG, "Run method for %s linked at %p\n", plugin->init_argv[0], run_fnc_ptr);
  EEF_log(LOG_DEBUG, "Running plugin %s\n", plugin->init_argv[0]);
  set_running_plugin(plugin);
  return run_fnc_ptr();
}

/**
 * run the plugin by name
 */
EES_PL_RC run_plugin_by_name(char* plugin_name){
  eef_plugindl_t * current_plugin;
  if((current_plugin = get_plugin(plugin_name))){
    if((run_fnc_ptr = (run_fnc_ptr_t)current_plugin->procs[RUNPROC])){
      EEF_log(LOG_DEBUG, "Run method for %s linked at %p\n", current_plugin->init_argv[0], run_fnc_ptr);
      EEF_log(LOG_DEBUG, "Running plugin %s\n", current_plugin->init_argv[0]);
      set_running_plugin(current_plugin);
      run_fnc_ptr();
      return EES_SUCCESS;
    }
  }
  return EES_FAILURE;
}

/**
 * Terminates all the plugins
 * Returns EES_SUCCESS if all plugins terminated successfully.
 */
EES_RC term_plugins(void){
  eef_plugindl_t *node = _plugin_list;
  EES_RC retval = EES_SUCCESS;
  while(node != NULL){
    if(term_plugin(node->name) == EES_PL_FAILURE){
      retval = EES_FAILURE;
    }
    node = node->next;
  }
  return retval;
}

/**
 * Calls the plugin's term method
 */
EES_PL_RC term_plugin(char* plugin_name){
  eef_plugindl_t * current_plugin;
  EES_RC           retval = EES_SUCCESS;
  if((current_plugin = get_plugin(plugin_name))){
    if((term_fnc_ptr = (term_fnc_ptr_t)current_plugin->procs[TERMPROC])){
      EEF_log(LOG_DEBUG, "Term method for %s linked at %p\n", plugin_name, term_fnc_ptr);
      EEF_log(LOG_DEBUG, "Terminating plugin %s\n", plugin_name);
      set_running_plugin(current_plugin);
      if(term_fnc_ptr() != EES_PL_SUCCESS){
        retval = EES_FAILURE;
      }
    }
  }
  return retval;
}

/**
 * Stops the plugin manager, calling clean_plugin_list
 */
EES_RC stop_plugin_manager(){
  /*EES_RC retval = EES_SUCCESS;*/
  if((term_plugins() == EES_SUCCESS) && (clean_plugin_list(_plugin_list) == EES_SUCCESS)){
    return EES_SUCCESS;
  }
  return EES_FAILURE;
}

/**
 * Cleans the list of plugins, unlinking all the dlopened modules and freeing the structs
 * \post The plugin list is empty
 */
EES_RC clean_plugin_list(eef_plugindl_t * list){
	eef_plugindl_t * plugin_entry   = list;
	eef_plugindl_t * plugin_next    = NULL;
  EES_RC           retval         = EES_SUCCESS;
  int              i;

  /* iterate plugin_list */
	while(plugin_entry!=NULL){
    if(plugin_entry->handle != NULL){
      /* Valgrind can't trace modules past their dlclose-ing. See: http://valgrind.org/docs/manual/faq.html#faq.unhelpful */
      if(!getenv("VALGRIND")){
        if((dlclose(plugin_entry->handle))){
          EEF_log(LOG_ERR, "dlclose error %s while cleaning up plugin list\n", dlerror());
          retval = EES_FAILURE;
        }
      } else {
        EEF_log(LOG_DEBUG, "Running in valgrind, not dlclose'ing plugins\n");
      }
    }

    /* free argv array */
    for(i = 0; i < plugin_entry->init_argc; i++){
      free(plugin_entry->init_argv[i]);
    }

    /* free current struct and move to next plugin */
    plugin_next = plugin_entry->next;
    free(plugin_entry);
    plugin_entry = plugin_next;
	}
	return retval;
}


/**
 * Prints the information from the plugin struct to the eef_log function with a specified debug level
 */
EES_RC print_eef_plugin(int debug_lvl, eef_plugindl_t *plugin){
  int i = 0;
  for(i = 1; i < plugin->init_argc; i++){
    EEF_log(debug_lvl, "plugin %s: argv[%i]            : %s\n", plugin->init_argv[0] , i, plugin->init_argv[i]);
  }
	return EES_SUCCESS;
}

/**
 * Internal method to keep track of which plugin is currently executing (used by AOS to determine ownership of data)
 */
eef_plugindl_t * get_running_plugin(){
  return _global_running_plugin;
}

char* get_running_plugin_name(void){
  char* name = NULL;
  if(!_is_threading){
    return _global_running_plugin_name;
  }

  /*  Try to get TLS */
  name = pthread_getspecific(_plugin_name_key);
  if(name == NULL){
    pthread_setspecific(_plugin_name_key, name);
  }
  return name;
}

/**
 * sets threading flag
 */
void plugin_manager_start_threading(void){
	_is_threading = 1;
}

/**
 * Internal method to keep track of which plugin is currently executing
 */
void set_running_plugin(eef_plugindl_t * plugin){
  if(!_is_threading){
    _global_running_plugin = plugin;
  } else {
    pthread_setspecific(_plugin_key, plugin);
  }
  return;
}

/**
 * Returns the name of the currently running plugin
 */
void set_running_plugin_name(char* plugin_name){
  if(!_is_threading){
    _global_running_plugin_name = plugin_name;
  } else {
    pthread_setspecific(_plugin_name_key, plugin_name);
  }
  return;
}

/**
 * Creates thread-local storage key
 */
void plugin_manager_make_key(void){
  pthread_key_create(&_plugin_key, NULL);
  pthread_key_create(&_plugin_name_key, NULL);
}

